# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import abc
import dataclasses
import functools

import gt4py._core.definitions as core_defs
import gt4py.next.common as common
import gt4py.storage.allocators as core_allocators
from gt4py.eve.extended_typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Final,
    Optional,
    Protocol,
    Sequence,
    TypeAlias,
    TypeIs,
    cast,
)


FieldLayoutMapper: TypeAlias = Callable[
    [Sequence[common.Dimension]], core_allocators.BufferLayoutMap
]


class FieldBufferAllocatorProtocol(Protocol[core_defs.DeviceTypeT]):
    """Protocol for buffer allocators used to allocate memory for fields with a given domain."""

    @property
    @abc.abstractmethod
    def __gt_device_type__(self) -> core_defs.DeviceTypeT: ...

    @abc.abstractmethod
    def __gt_allocate__(
        self,
        domain: common.Domain,
        dtype: core_defs.DType[core_defs.ScalarT],
        device_id: int = 0,
        aligned_index: Optional[Sequence[common.NamedIndex]] = None,  # absolute position
    ) -> core_allocators.TensorBuffer[core_defs.DeviceTypeT, core_defs.ScalarT]: ...


def is_field_allocator(obj: Any) -> TypeIs[FieldBufferAllocatorProtocol]:
    return hasattr(obj, "__gt_device_type__") and hasattr(obj, "__gt_allocate__")


def is_field_allocator_for(
    obj: Any, device: core_defs.DeviceTypeT
) -> TypeIs[FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]]:
    return is_field_allocator(obj) and obj.__gt_device_type__ is device


class FieldBufferAllocatorFactoryProtocol(Protocol[core_defs.DeviceTypeT]):
    """Protocol for device-specific buffer allocator factories for fields."""

    @property
    @abc.abstractmethod
    def __gt_allocator__(self) -> FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]: ...


def is_field_allocator_factory(obj: Any) -> TypeIs[FieldBufferAllocatorFactoryProtocol]:
    return hasattr(obj, "__gt_allocator__")


def is_field_allocator_factory_for(
    obj: Any, device: core_defs.DeviceTypeT
) -> TypeIs[FieldBufferAllocatorFactoryProtocol[core_defs.DeviceTypeT]]:
    return is_field_allocator_factory(obj) and obj.__gt_allocator__.__gt_device_type__ is device


FieldBufferAllocationUtil = (
    FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]
    | FieldBufferAllocatorFactoryProtocol[core_defs.DeviceTypeT]
)


def is_field_allocation_tool(obj: Any) -> TypeIs[FieldBufferAllocationUtil]:
    return is_field_allocator(obj) or is_field_allocator_factory(obj)


def is_field_allocation_tool_for(
    obj: Any, device: core_defs.DeviceTypeT
) -> TypeIs[FieldBufferAllocationUtil]:
    return is_field_allocator_for(obj, device) or is_field_allocator_factory_for(obj, device)


def _absolute_to_relative_index(
    indices: Sequence[common.NamedIndex], domain: common.Domain
) -> Sequence[int]:
    """Convert absolute indices to relative indices based on the domain's dimensions.

    Indices for which the dimension isn't represented in the domain are ignored,
    and the resulting relative index sequence is ordered according to the domain's dimensions.
    """

    dim_to_index = {named_idx.dim: int(named_idx.value) for named_idx in indices}
    result = []
    for named_range in domain:
        dim = named_range.dim
        abs_idx = dim_to_index.get(dim, named_range.unit_range.start)
        if abs_idx < named_range.unit_range.start or abs_idx >= named_range.unit_range.stop:
            raise ValueError(
                f"Absolute index for dimension '{dim}' is {abs_idx}, "
                f"which is outside the domain range [{named_range.unit_range.start}, {named_range.unit_range.stop})."
            )
        result.append(abs_idx - named_range.unit_range.start)
    return result


@dataclasses.dataclass(frozen=True)
class BaseFieldBufferAllocator(FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]):
    """Parametrizable field buffer allocator base class."""

    device_type: core_defs.DeviceTypeT
    array_utils: core_allocators.ArrayUtils
    layout_mapper: FieldLayoutMapper
    byte_alignment: int

    @functools.cached_property
    def buffer_allocator(self) -> core_allocators.BufferAllocator:
        return core_allocators.NDArrayBufferAllocator(self.device_type, self.array_utils)

    @property
    def __gt_device_type__(self) -> core_defs.DeviceTypeT:
        return self.device_type

    def __gt_allocate__(
        self,
        domain: common.Domain,
        dtype: core_defs.DType[core_defs.ScalarT],
        device_id: int = 0,
        aligned_index: Optional[Sequence[common.NamedIndex]] = None,  # absolute position
    ) -> core_allocators.TensorBuffer[core_defs.DeviceTypeT, core_defs.ScalarT]:
        shape = domain.shape
        layout_map = self.layout_mapper(domain.dims)

        relative_aligned_index = (
            _absolute_to_relative_index(aligned_index, domain) if aligned_index else None
        )

        return self.buffer_allocator.allocate(
            shape, dtype, device_id, layout_map, self.byte_alignment, relative_aligned_index
        )


if TYPE_CHECKING:
    __TensorFieldAllocatorAsFieldAllocatorInterfaceT: type[FieldBufferAllocatorProtocol] = (
        BaseFieldBufferAllocator
    )


def horizontal_first_layout_mapper(
    dims: Sequence[common.Dimension],
) -> core_allocators.BufferLayoutMap:
    """Map dimensions to a buffer layout making horizontal dims change the slowest (i.e. larger strides)."""

    def pos_of_kind(kind: common.DimensionKind) -> list[int]:
        return [i for i, dim in enumerate(dims) if dim.kind == kind]

    horizontals = pos_of_kind(common.DimensionKind.HORIZONTAL)
    verticals = pos_of_kind(common.DimensionKind.VERTICAL)
    locals_ = pos_of_kind(common.DimensionKind.LOCAL)

    layout_map = [0] * len(dims)
    for i, pos in enumerate(horizontals + verticals + locals_):
        layout_map[pos] = len(dims) - 1 - i

    valid_layout_map = tuple(layout_map)
    assert core_allocators.is_valid_layout_map(valid_layout_map)

    return valid_layout_map


if TYPE_CHECKING:
    __horizontal_first_layout_mapper: FieldLayoutMapper = horizontal_first_layout_mapper


#: Registry of default allocators for each device type.
device_allocators: dict[core_defs.DeviceType, FieldBufferAllocatorProtocol] = {}


class StandardCPUFieldBufferAllocator(BaseFieldBufferAllocator[core_defs.CPUDeviceTyping]):
    """A field buffer allocator for CPU devices that uses a horizontal-first layout mapper and 64-byte alignment."""

    def __init__(self) -> None:
        super().__init__(
            device_type=core_defs.DeviceType.CPU,
            array_utils=core_allocators.numpy_array_utils,
            layout_mapper=horizontal_first_layout_mapper,
            byte_alignment=64,
        )


device_allocators[core_defs.DeviceType.CPU] = StandardCPUFieldBufferAllocator()


assert is_field_allocator(device_allocators[core_defs.DeviceType.CPU])


@dataclasses.dataclass(frozen=True)
class InvalidFieldBufferAllocator(FieldBufferAllocatorProtocol[core_defs.DeviceTypeT]):
    """A field buffer allocator that always raises an exception."""

    device_type: core_defs.DeviceTypeT
    exception: Exception

    @property
    def __gt_device_type__(self) -> core_defs.DeviceTypeT:
        return self.device_type

    def __gt_allocate__(
        self,
        domain: common.Domain,
        dtype: core_defs.DType[core_defs.ScalarT],
        device_id: int = 0,
        aligned_index: Optional[Sequence[common.NamedIndex]] = None,  # absolute position
    ) -> core_allocators.TensorBuffer[core_defs.DeviceTypeT, core_defs.ScalarT]:
        raise self.exception


if core_defs.CUPY_DEVICE_TYPE is not None:
    assert isinstance(core_allocators.cupy_array_utils, core_allocators.ArrayUtils)
    cupy_array_utils = core_allocators.cupy_array_utils

    if core_defs.CUPY_DEVICE_TYPE is core_defs.DeviceType.CUDA:

        class CUDAFieldBufferAllocator(BaseFieldBufferAllocator[core_defs.CUDADeviceTyping]):
            def __init__(self) -> None:
                super().__init__(
                    device_type=core_defs.DeviceType.CUDA,
                    array_utils=cupy_array_utils,
                    layout_mapper=horizontal_first_layout_mapper,
                    byte_alignment=128,
                )

        device_allocators[core_defs.DeviceType.CUDA] = CUDAFieldBufferAllocator()

    else:

        class ROCMFieldBufferAllocator(BaseFieldBufferAllocator[core_defs.ROCMDeviceTyping]):
            def __init__(self) -> None:
                super().__init__(
                    device_type=core_defs.DeviceType.ROCM,
                    array_utils=cupy_array_utils,
                    layout_mapper=horizontal_first_layout_mapper,
                    byte_alignment=128,
                )

        device_allocators[core_defs.DeviceType.ROCM] = ROCMFieldBufferAllocator()

else:

    class InvalidGPUFieldBufferAllocator(InvalidFieldBufferAllocator[core_defs.CUDADeviceTyping]):
        def __init__(self) -> None:
            super().__init__(
                device_type=core_defs.DeviceType.CUDA,
                exception=RuntimeError("Missing `cupy` dependency for GPU allocation"),
            )


StandardGPUFieldBufferAllocator: Final[type[FieldBufferAllocatorProtocol]] = cast(
    type[FieldBufferAllocatorProtocol],
    type(device_allocators[core_defs.CUPY_DEVICE_TYPE])
    if core_defs.CUPY_DEVICE_TYPE
    else InvalidGPUFieldBufferAllocator,
)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""Package metadata: version, authors, license and copyright."""

import pathlib
from typing import Any, Final

from packaging import version as pkg_version


__all__ = [
    "__author__",
    "__copyright__",
    "__license__",
    "__version__",
    "__version_info__",
]


__author__: Final = "ETH Zurich and individual contributors"
__copyright__: Final = "Copyright (c) 2014-2024 ETH Zurich"
__license__: Final = "BSD-3-Clause"

__version__: str
__version_info__: pkg_version.Version


#: This value should be overwritten with the actual version string at build
#: time by the `onbuild` hook of versioningit. If the hook is not run for
#: whatever reason, the current value defined here would be used as fallback.
#: Therefore, for consistency, the version hard-coded here should be kept in
#: sync with the `tool.versioningit.default-version` field in `pyproject.toml`.
on_build_version: Final = "1.1.8"

_cached_version_data: tuple[str, pkg_version.Version] | None = None
_dir: list[str] | None = None


def _get_version_from_versioningit(path: pathlib.Path | str) -> str:
    import os

    import versioningit

    devnull = open(os.devnull, "w")
    try:
        old_stdout_fd = os.dup(1)  # duplicate current stdout fd
        os.dup2(devnull.fileno(), 1)  # replace fd 1 with /dev/null
        old_stderr_fd = os.dup(2)
        os.dup2(devnull.fileno(), 2)

        version = versioningit.get_version(path)

    finally:
        os.dup2(old_stdout_fd, 1)
        os.close(old_stdout_fd)
        os.dup2(old_stderr_fd, 2)
        os.close(old_stderr_fd)

    return version


def _inspect_version() -> tuple[str, pkg_version.Version]:
    global _cached_version_data

    if _cached_version_data is not None:
        return _cached_version_data

    import importlib.metadata

    if dist := importlib.metadata.distribution("gt4py"):
        version: str = dist.version  # static installation version

        if contents := dist.read_text("direct_url.json"):
            # This branch should only be executed in editable installs.
            # In this case, the version reported by `gt4py.__version__`
            # is directly computed from the status of the `git` repository
            # with the source code, if available, which might differ from the
            # version reported by `importlib.metadata.version("gt4py")`
            # (and `$ pip show gt4py`).
            try:
                import json

                url_data = json.loads(contents)
                if url_data["dir_info"]["editable"] is not True:
                    raise ValueError("Not an editable install")
                if not url_data["url"].startswith("file://"):
                    raise ValueError("Not a local install")

                src_path = url_data["url"][7:]
                version = _get_version_from_versioningit(src_path)
            except Exception:
                # There is something wrong in the current editable installation.
                # Fallback to the static version, but don't cache the result as the
                # final version data, since the package is installed in editable mode.
                pass

            version_info = pkg_version.parse(version)

            return version, version_info

        else:
            # If the package is not installed in editable mode, the version
            # is always correctly reported by `importlib.metadata.version("gt4py")`.
            _cached_version_data = version, pkg_version.parse(version)

            return _cached_version_data

    else:
        # This branch is a weird case: since `importlib.metadata`
        # couldn't find the distribution info, it should only be executed
        # if the package is not properly installed but it runs anyway.
        # For example, when running directly from sources by adding the
        # project root to the `sys.path`.
        try:
            version = _get_version_from_versioningit(pathlib.Path(__file__).parent.resolve())
        except Exception:
            # Fallback to the on-build version, if everything else fails.
            version = on_build_version

        version_info = pkg_version.parse(version)

        return version, version_info


def __getattr__(name: str) -> Any:
    match name:
        case "__version__":
            return _inspect_version()[0]
        case "__version_info__":
            return _inspect_version()[1]
        case _:
            raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def __dir__() -> list[str]:
    global _dir
    if _dir is None:
        import sys

        # Add virtual attributes (e.g. __version__, __version_info__) to the list
        annotations = globals().get("__annotations__", {})
        _dir = list(set(annotations.keys()) | vars(sys.modules[__name__]).keys())

    return _dir



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
Definitions of the stages of the GT4Py frontend.

Classes in this module contain different forms of field operator and program
definitions, which are used as input or output of the different stages of
the frontend.

All classes containing a definition of a GT4Py computation in any form use the
`Def` suffix. Definitions containing actual Python functions whose source code
should be interpreted as GT4Py embedded domain-specific language have `DSL` in
their name. Definitions containing definitions as an AST of one the internal GT4Py
dialects contain `AST`.
"""

from __future__ import annotations

import collections.abc
import dataclasses
import functools
import hashlib
import types
import typing
from typing import Any, Optional, TypeVar

import xxhash

from gt4py.eve import extended_typing as xtyping
from gt4py.next import common
from gt4py.next.ffront import field_operator_ast as foast, program_ast as past, source_utils
from gt4py.next.otf import arguments, toolchain


@dataclasses.dataclass(frozen=True)
class DSLFieldOperatorDef:
    definition: types.FunctionType
    node_class: type[foast.OperatorNode] = foast.FieldOperator
    attributes: dict[str, Any] = dataclasses.field(default_factory=dict)
    grid_type: Optional[common.GridType] = None
    debug: bool = False


ConcreteDSLFieldOperatorDef: typing.TypeAlias = toolchain.ConcreteArtifact[
    DSLFieldOperatorDef, arguments.CompileTimeArgs
]


@dataclasses.dataclass(frozen=True)
class FOASTOperatorDef:
    foast_node: foast.OperatorNode
    closure_vars: dict[str, Any]
    grid_type: Optional[common.GridType] = None
    attributes: dict[str, Any] = dataclasses.field(default_factory=dict)
    debug: bool = False


ConcreteFOASTOperatorDef: typing.TypeAlias = toolchain.ConcreteArtifact[
    FOASTOperatorDef, arguments.CompileTimeArgs
]


@dataclasses.dataclass(frozen=True)
class DSLProgramDef:
    definition: types.FunctionType
    grid_type: Optional[common.GridType] = None
    debug: bool = False


ConcreteDSLProgramDef: typing.TypeAlias = toolchain.ConcreteArtifact[
    DSLProgramDef, arguments.CompileTimeArgs
]


@dataclasses.dataclass(frozen=True)
class PASTProgramDef:
    past_node: past.Program
    closure_vars: dict[str, Any]
    grid_type: Optional[common.GridType] = None
    debug: bool = False


ConcretePASTProgramDef: typing.TypeAlias = toolchain.ConcreteArtifact[
    PASTProgramDef, arguments.CompileTimeArgs
]

DSLDefinition = DSLFieldOperatorDef | DSLProgramDef
DSLDefinitionT = TypeVar("DSLDefinitionT", DSLFieldOperatorDef, DSLProgramDef)


def fingerprint_stage(obj: Any, algorithm: Optional[str | xtyping.HashlibAlgorithm] = None) -> str:
    hasher: xtyping.HashlibAlgorithm
    if not algorithm:
        hasher = xxhash.xxh64()  # type: ignore[assignment]  # fixing this requires https://github.com/ifduyue/python-xxhash/issues/104
    elif isinstance(algorithm, str):
        hasher = hashlib.new(algorithm)
    else:
        hasher = algorithm

    add_content_to_fingerprint(obj, hasher)
    return hasher.hexdigest()


@functools.singledispatch
def add_content_to_fingerprint(obj: Any, hasher: xtyping.HashlibAlgorithm) -> None:
    hasher.update(str(obj).encode())


for t in (str, int):
    add_content_to_fingerprint.register(t, add_content_to_fingerprint.registry[object])


@add_content_to_fingerprint.register(DSLFieldOperatorDef)
@add_content_to_fingerprint.register(FOASTOperatorDef)
@add_content_to_fingerprint.register(DSLProgramDef)
@add_content_to_fingerprint.register(PASTProgramDef)
@add_content_to_fingerprint.register(toolchain.ConcreteArtifact)
@add_content_to_fingerprint.register(arguments.CompileTimeArgs)
def add_stage_to_fingerprint(obj: Any, hasher: xtyping.HashlibAlgorithm) -> None:
    add_content_to_fingerprint(obj.__class__, hasher)
    for field in dataclasses.fields(obj):
        add_content_to_fingerprint(getattr(obj, field.name), hasher)


def add_jit_args_id_to_fingerprint(
    obj: arguments.JITArgs, hasher: xtyping.HashlibAlgorithm
) -> None:
    add_content_to_fingerprint(str(id(obj)), hasher)


@add_content_to_fingerprint.register
def add_func_to_fingerprint(obj: types.FunctionType, hasher: xtyping.HashlibAlgorithm) -> None:
    sourcedef = source_utils.SourceDefinition.from_function(obj)
    for item in sourcedef:
        add_content_to_fingerprint(item, hasher)

    closure_vars = source_utils.get_closure_vars_from_function(obj)
    for item in sorted(closure_vars.items(), key=lambda x: x[0]):
        add_content_to_fingerprint(item, hasher)


@add_content_to_fingerprint.register
def add_dict_to_fingerprint(obj: dict, hasher: xtyping.HashlibAlgorithm) -> None:
    # just a small helper to additionally allow sorting types (by just using their name)
    def key_function(key: Any) -> Any:
        if isinstance(key, type):
            return key.__module__, key.__qualname__
        return key

    for key in sorted(obj.keys(), key=key_function):
        add_content_to_fingerprint(key, hasher)
        add_content_to_fingerprint(obj[key], hasher)


@add_content_to_fingerprint.register
def add_type_to_fingerprint(obj: type, hasher: xtyping.HashlibAlgorithm) -> None:
    hasher.update(obj.__name__.encode())


@add_content_to_fingerprint.register
def add_sequence_to_fingerprint(
    obj: collections.abc.Iterable, hasher: xtyping.HashlibAlgorithm
) -> None:
    for item in obj:
        add_content_to_fingerprint(item, hasher)


@add_content_to_fingerprint.register
def add_foast_located_node_to_fingerprint(
    obj: foast.LocatedNode, hasher: xtyping.HashlibAlgorithm
) -> None:
    add_content_to_fingerprint(obj.location, hasher)
    add_content_to_fingerprint(str(obj), hasher)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import ast
from typing import Any, Optional


class StringifyAnnotationsPass(ast.NodeTransformer):
    """
    AST pass transforming type annotations to string :class:`ast.Constant`.

    Mutates the python AST replacing arbitrary AST subtrees in the annotation
    fields of the supported nodes with its string representation
    using :func:`ast.unparse()`.


    Example
    -------
    >>> import ast, inspect
    >>> from typing import Union

    >>> def foo(a: Union[int, float]) -> float:
    ...     tmp: Union[int, float] = a + 1
    ...     result: float = float(tmp)
    ...     return result

    >>> ast_node = ast.parse(inspect.getsource(foo))
    >>> print(
    ...     f"_ {ast_node.body[0].args.args[0].annotation}"
    ... )  # first character of output can not be ignored
    _ ...Subscript...

    >>> ast_node = StringifyAnnotationsPass.apply(ast_node)
    >>> print(f"_ {ast_node.body[0].args.args[0].annotation}")
    _ ...Constant...
    """

    @classmethod
    def apply(cls, node: ast.AST) -> ast.AST:
        return cls().visit(node)

    def visit_arg(self, node: ast.arg) -> ast.arg:
        node.annotation = self._stringify_annotation(node.annotation)
        result = self.generic_visit(node)
        assert isinstance(result, ast.arg)
        return result

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AnnAssign:
        node.annotation = self._stringify_annotation(node.annotation)
        result = self.generic_visit(node)
        assert isinstance(result, ast.AnnAssign)
        return result

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        node.returns = self._stringify_annotation(node.returns)
        result = self.generic_visit(node)
        assert isinstance(result, ast.FunctionDef)
        return result

    @staticmethod
    def _stringify_annotation(node: Optional[ast.AST]) -> Any:
        return ast.Constant(value=ast.unparse(node), kind=None) if node else node



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import atexit
import collections
import contextlib
import contextvars
import dataclasses
import functools
import itertools
import json
import numbers
import operator
import pathlib
import sys
import time
import types
import typing
from collections.abc import Callable, Mapping
from typing import ClassVar, TypeAlias

import numpy as np

from gt4py.eve import extended_typing as xtyping, utils
from gt4py.eve.extended_typing import Any, Final
from gt4py.next import config
from gt4py.next.otf import arguments


_NO_KEY_SET_MARKER_: Final[str] = sys.intern("@_NO_KEY_SET_MARKER_@")


# Common metric names
COMPUTE_METRIC: Final[str] = sys.intern("compute")
TOTAL_METRIC: Final[str] = sys.intern("total")


# Metric collection levels
DISABLED: Final[int] = 0
MINIMAL: Final[int] = 1
PERFORMANCE: Final[int] = 10
INFO: Final[int] = 30
VERBOSE: Final[int] = 50
ALL: Final[int] = 100


def is_any_level_enabled() -> bool:
    """Check if any metrics collection level is enabled."""
    return config.COLLECT_METRICS_LEVEL > DISABLED


def is_level_enabled(level: int) -> bool:
    """Check if a given metrics collection level is enabled."""
    return config.COLLECT_METRICS_LEVEL >= level


def get_current_level() -> int:
    """Retrieve the current metrics collection level (from the configuration module)."""
    return config.COLLECT_METRICS_LEVEL


@dataclasses.dataclass(frozen=True)
class Metric:
    """
    A class to collect and analyze samples of a metric.

    Examples:
        >>> metric = Metric(name="execution_time")
        >>> metric.add_sample(0.1)
        >>> metric.add_sample(0.2)
        >>> print(f"{metric.mean:.2}")
        0.15
        >>> print(metric)
        1.50000e-01 +/- 7.07107e-02
    """

    name: str | None = None
    samples: list[float] = dataclasses.field(default_factory=list)

    def __post_init__(self) -> None:
        if self.name:
            object.__setattr__(self, "name", sys.intern(self.name))

    @property
    def mean(self) -> np.floating:
        if len(self.samples) == 0:
            raise ValueError("Cannot compute mean of empty sample list.")
        return np.mean(self.samples)

    @property
    def std(self) -> np.floating:
        if len(self.samples) == 0:
            raise ValueError("Cannot compute std of empty sample list.")
        return np.std(self.samples, ddof=1)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, values={self.samples})"

    def __str__(self) -> str:
        return f"{self.mean:.5e} +/- {self.std:.5e}"

    def add_sample(self, sample: float) -> None:
        self.samples.append(sample)


class MetricsCollection(utils.CustomDefaultDictBase[str, Metric]):
    """
    A collection of metrics, organized as a mapping from metric names to `Metric` objects.

    Empty `Metric` instances are created automatically when accessing keys
    that do not exist.

    Example:
        >>> metrics = MetricsCollection()
        >>> metrics["execution_time"].add_sample(0.1)
        >>> metrics["execution_time"].add_sample(0.2)
        >>> metrics["execution_time"].samples
        [0.1, 0.2]
    """

    __slots__ = ()

    def value_factory(self, key: str) -> Metric:
        assert isinstance(key, str)
        return Metric(name=key)


@dataclasses.dataclass(slots=True)
class Source:  # type: ignore[misc]  # Mypy bug fixed by: https://github.com/python/mypy/pull/20573
    """A source of metrics, typically associated with a program."""

    metadata: dict[str, Any] = dataclasses.field(default_factory=dict)
    metrics: MetricsCollection = dataclasses.field(default_factory=MetricsCollection)


#: Global store for all measurements.
sources: collections.defaultdict[str, Source] = collections.defaultdict(Source)

# Context variable storing the active source key.
_source_key_cvar: contextvars.ContextVar[str] = contextvars.ContextVar("source_key")


def is_current_source_key_set() -> bool:
    """Check if there is a source key set for metrics collection."""
    return _source_key_cvar.get(_NO_KEY_SET_MARKER_) is not _NO_KEY_SET_MARKER_


def set_current_source_key(key: str) -> None:
    """
    Set the current source key for metrics collection.

    It must be called only when no source key is set (or the same key is already set).
    """
    assert _source_key_cvar.get(_NO_KEY_SET_MARKER_) in {key, _NO_KEY_SET_MARKER_}, (
        "A different source key has been already set."
    )
    _source_key_cvar.set(key)


def get_current_source_key() -> str:
    """Retrieve the current source key for metrics collection (it must be set)."""
    return _source_key_cvar.get()


def get_current_source() -> Source:
    """Retrieve the active metrics collection source (a valid source key must be set)."""
    return sources[_source_key_cvar.get()]


def add_sample_to_current_source(metric_name: str, sample: float) -> None:
    """Add a sample to a metric in the current source (a valid source key must be set)."""
    return sources[_source_key_cvar.get()].metrics[metric_name].add_sample(sample)


@dataclasses.dataclass(slots=True)
class SourceKeySetterAtEnter(contextlib.AbstractContextManager):  # type: ignore[misc]  # Mypy bug fixed by: https://github.com/python/mypy/pull/20573
    """
    A context manager to handle metrics collection source keys.

    When entering this context manager it saves the current source key
    for metrics collection and sets the new source key if provided, or
    a default marker indicating no key is set. Upon exiting the context,
    it does not reset the source key itself, but instead relies on another
    component (usually an outer `program_call_context`) to reset the
    source key later.
    """

    key: str | None = None
    previous_cvar_token: contextvars.Token | None = dataclasses.field(init=False)

    # This class is implemented as a context manager class instead of a generator
    # function with `@contextlib.contextmanager` to avoid the extra overhead
    # of renewing the generator inside `contextlib.contextmanager`.
    def __enter__(self) -> None:
        if is_any_level_enabled():
            self.previous_cvar_token = _source_key_cvar.set(self.key or _NO_KEY_SET_MARKER_)
        else:
            self.previous_cvar_token = None

    def __exit__(
        self,
        exc_type_: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        pass


@dataclasses.dataclass(slots=True)
class SourceKeyContextManager(SourceKeySetterAtEnter):  # type: ignore[misc]  # Mypy bug fixed by: https://github.com/python/mypy/pull/20573
    """
    Another context manager to handle metrics collection source keys.

    It works like `SourceKeySetterAtEnter` at enter, but it additionally
    resets the source key to its previous state upon exiting the context.
    """

    def __exit__(
        self,
        exc_type_: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        if self.previous_cvar_token is not None:
            _source_key_cvar.reset(self.previous_cvar_token)


metrics_source_key_setter: TypeAlias = SourceKeySetterAtEnter
metrics_source_key_context: TypeAlias = SourceKeyContextManager


@dataclasses.dataclass(slots=True)
class BaseMetricsCollector(contextlib.AbstractContextManager):  # type: ignore[misc]  # Mypy bug fixed by: https://github.com/python/mypy/pull/20573
    """
    A context manager to handle metrics collection.

    This is a base class for creating metrics collectors that measure
    specific metrics during the execution of a code block. It provides
    a convenient interface for managing the lifecycle of metrics collection.

    Subclasses need to define the `level` and `metric_name` attributes, and,
    optionally override the methods for collecting counters and computing
    the metric. This class offers a simple way to customize this class variables
    accepting them as keyword arguments when creating the subclass.
    """

    def __init_subclass__(
        cls: type[BaseMetricsCollector],
        *,
        level: int,
        metric_name: str,
        collect_enter_counter: Callable[[], float] | None = None,
        collect_exit_counter: Callable[[], float] | None = None,
        compute_metric: Callable[[float, float], float] | None = None,
        **kwargs: Any,
    ) -> None:
        super(BaseMetricsCollector, cls).__init_subclass__(**kwargs)
        cls.level = level
        cls.metric_name = sys.intern(metric_name)
        if collect_enter_counter is not None:
            cls.collect_enter_counter = staticmethod(collect_enter_counter)
        if collect_exit_counter is not None:
            cls.collect_exit_counter = staticmethod(collect_exit_counter)
        if compute_metric is not None:
            cls.compute_metric = staticmethod(compute_metric)

    # Subclass must define these class variables
    level: ClassVar[int]
    metric_name: ClassVar[str]

    # Default implementations for these methods can be overridden by subclasses
    collect_enter_counter: ClassVar[Callable[[], float]] = staticmethod(time.perf_counter)
    collect_exit_counter: ClassVar[Callable[[], float]] = staticmethod(time.perf_counter)
    compute_metric: ClassVar[Callable[[float, float], float]] = staticmethod(operator.sub)
    #: compute_metric(exit_counter, enter_counter) -> metric_value

    # Instance state
    key: str | None = None
    previous_cvar_token: contextvars.Token | None = dataclasses.field(init=False)
    enter_counter: float | None = dataclasses.field(init=False)

    def __enter__(self) -> None:
        if is_level_enabled(self.level):
            self.previous_cvar_token = _source_key_cvar.set(self.key or _NO_KEY_SET_MARKER_)
            self.enter_counter = self.collect_enter_counter()  # type: ignore[misc] # mypy doesn't understand that this is a staticmethod
        else:
            self.previous_cvar_token = None

    def __exit__(
        self,
        exc_type_: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: types.TracebackType | None,
    ) -> None:
        if self.previous_cvar_token is not None:
            assert is_current_source_key_set() is True
            assert hasattr(self, "enter_counter") and self.enter_counter is not None
            sources[_source_key_cvar.get()].metrics[self.metric_name].add_sample(
                self.compute_metric(self.collect_exit_counter(), self.enter_counter)  # type: ignore[call-arg,misc] # mypy doesn't understand that this is a staticmethod
            )
            _source_key_cvar.reset(self.previous_cvar_token)


@functools.cache
def make_collector(
    level: int,
    metric_name: str,
    *,
    collect_enter_counter: Callable[[], float] | None = None,
    collect_exit_counter: Callable[[], float] | None = None,
    compute_metric: Callable[[float, float], float] | None = None,
) -> type[BaseMetricsCollector]:
    """
    Create a custom metrics collector class.

    This function generates a new subclass of `BaseMetricsCollector` with
    the specified configuration for metrics collection.

    Args:
        level: The metrics collection level.
        metric_name: The name of the metric to be collected.
        collect_enter_counter: Optional function to collect the enter counter.
        collect_exit_counter: Optional function to collect the exit counter.
        compute_metric: Optional function to compute the metric from the counters.

    Returns:
        A new subclass of `BaseMetricsCollector` configured with the provided parameters.
    """

    return types.new_class(
        f"AutoMetricsCollectorFor_{metric_name}",
        bases=(BaseMetricsCollector,),
        kwds=dict(
            level=level,
            metric_name=metric_name,
            collect_enter_counter=collect_enter_counter,
            collect_exit_counter=collect_exit_counter,
            compute_metric=compute_metric,
        ),
        exec_body=None,
    )


def dumps(metric_sources: Mapping[str, Source] | None = None) -> str:
    """
    Format the metrics in the collection store as a string table.

    This function generates a formatted string representation of the metrics
    in the collection store. Each row represents a program, and each column
    represents a metric, showing both the mean value and standard deviation.

    If no explicit collection store is provided, uses the global `program_metrics`.
    """
    if metric_sources is None:
        metric_sources = typing.cast(Mapping[str, Source], sources)
    assert metric_sources is not None

    source_names: list[str] = [*sources.keys()]
    metric_names: list[str] = [
        *dict.fromkeys(
            itertools.chain.from_iterable(
                source.metrics.keys() for source in metric_sources.values()
            )
        ).keys()
    ]
    title_program = "program"
    title_cols = max(len(k) for k in (*metric_names, title_program)) + 1
    width = max([10, *(len(name) for name in (*metric_names, *source_names))])
    titles = (f"{name:<{width}}  {'+/-':<{width}}" for name in metric_names)
    header = f"{title_program:{title_cols}}  {'  '.join(titles)}"

    rows = []
    for name, source in metric_sources.items():
        cells = []
        for metric_name in metric_names:
            if metric_name in source.metrics:
                cells.append(
                    f"{source.metrics[metric_name].mean:<.4e}  {source.metrics[metric_name].std:<.4e}"
                )
            else:
                cells.append(f"{'N/A':<{width}} {'N/A':<{width}}")
        rows.append(f"{name:{title_cols}}  {'  '.join(cells)}")

    return str.join("\n", ["", header, *rows, ""])


def dump(filename: str | pathlib.Path, metric_sources: Mapping[str, Source] | None = None) -> None:
    pathlib.Path(filename).write_text(dumps(metric_sources))


def dumps_json(metric_sources: Mapping[str, Source] | None = None) -> str:
    """
    Export metric sources as a JSON string.

    If no explicit metric sources mapping is provided, the global `sources`
    will be used.

    Note:

        The exported JSON snippets have the following structure:

        ```json
        {
            source_key: {
                'metadata': {'key': value},
                'metrics': {'metric_name': [samples]}
            }
        }
        ```
    """
    if metric_sources is None:
        metric_sources = typing.cast(Mapping[str, Source], sources)
    assert metric_sources is not None

    def default_json_encoder(obj: object) -> object:
        match obj:
            case Source() as src:
                return {
                    "metadata": src.metadata,
                    "metrics": {name: metric.samples for name, metric in src.metrics.items()},
                }
            case arguments.StaticArg() as arg:
                return arg.value
            case xtyping.DataclassABC():
                return dataclasses.asdict(obj)
            case numbers.Integral() as i:
                return int(i)
            case numbers.Real() as r:
                return float(r)

        try:
            return str(obj)
        except Exception:
            raise TypeError(
                f"Object of type {type(obj).__name__} is not JSON serializable"
            ) from None

    return json.dumps(metric_sources, default=default_json_encoder)


def dump_json(
    filename: str | pathlib.Path, metric_sources: Mapping[str, Source] | None = None
) -> None:
    pathlib.Path(filename).write_text(dumps_json(metric_sources))


# Handler registration to automatically dump metrics at program exit if
# the corresponding configuration flag is set.
def _dump_metrics_at_exit() -> None:
    """Dump collected metrics to a file at program exit if required."""

    # It is assumed that 'gt4py.next.config' is still alive at this point
    if config.DUMP_METRICS_AT_EXIT and (is_any_level_enabled() or sources):
        try:
            pathlib.Path(config.DUMP_METRICS_AT_EXIT).write_text(dumps_json())
            print(
                f"--- atexit: GT4Py performance metrics saved at {config.DUMP_METRICS_AT_EXIT} ---",
                file=sys.stderr,
            )
        except Exception as e:
            print(
                f"--- atexit: ERROR: Failed to automatically save GT4Py performance metrics: ---\n{e}",
                file=sys.stderr,
            )


atexit.register(_dump_metrics_at_exit)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

from gt4py.cartesian import backend
from gt4py.cartesian.backend import python_common as py_common
from gt4py.cartesian.gtc import passes
from gt4py.cartesian.gtc.debug.debug_codegen import DebugCodeGen
from gt4py.cartesian.gtc.gtir_to_oir import GTIRToOIR
from gt4py.cartesian.gtc.passes import oir_optimizations
from gt4py.eve import codegen
from gt4py.storage import layout


if TYPE_CHECKING:
    from gt4py.cartesian.stencil_object import StencilObject


@backend.register
class DebugBackend(backend.BaseBackend):
    """Debug backend using plain python loops."""

    name = "debug"
    options: ClassVar[dict[str, Any]] = {
        "oir_pipeline": {"versioning": True, "type": passes.OirPipeline},
        "ignore_np_errstate": {"versioning": True, "type": bool},
    }
    storage_info = layout.NaiveCPULayout
    languages: ClassVar[dict[str, Any]] = {"computation": "python", "bindings": ["python"]}
    MODULE_GENERATOR_CLASS = py_common.PythonModuleGenerator

    def _generate_computation(self) -> dict[str, str | dict]:
        oir = GTIRToOIR().visit(self.builder.gtir)
        oir = oir_optimizations.HorizontalExecutionMerging().visit(oir)
        oir = oir_optimizations.LocalTemporariesToScalars().visit(oir)
        source_code = DebugCodeGen().visit(oir)

        if self.builder.options.format_source:
            source_code = codegen.format_source("python", source_code)

        caching = self.builder.caching
        computation_name = f"{caching.module_prefix}computation{caching.module_postfix}.py"
        return {computation_name: source_code}

    def generate(self) -> type[StencilObject]:
        self.check_options(self.builder.options)
        src_dir = self.builder.module_path.parent

        if not self.builder.options._impl_opts.get("disable-code-generation", False):
            py_common.recursive_write(src_dir, self._generate_computation())

        return self.make_module()



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

import copy
import dataclasses
import functools
import warnings
from typing import Any, Collection, Literal, Mapping, Optional, Sequence, TypeAlias, Union, overload

import dace
from dace import (
    data as dace_data,
    properties as dace_properties,
    subsets as dace_sbs,
    symbolic as dace_sym,
    transformation as dace_transformation,
)
from dace.sdfg import graph as dace_graph, nodes as dace_nodes
from dace.transformation.passes import analysis as dace_analysis

from gt4py.next import config as gtx_config
from gt4py.next.program_processors.runners.dace import transformations as gtx_transformations


@dace_properties.make_properties
class ConcatWhereCopyToMap(dace_transformation.SingleStateTransformation):
    """Replaces `concat_where` nodes with Tasklets that access the source data directly.

    This is essentially a wrapper around the `gt_replace_concat_where_node()`, for
    more information see there.
    There is also `gt_apply_concat_where_replacement_on_sdfg()`, it is recommended
    to use, as it guarantees stable ordering.

    Args:
        single_use_data: Single use data, if not provided a scan will be performed.
        tag: Used to mangle names of data descriptors, see `gt_replace_concat_where_node()`
            for more.
    """

    node_a1 = dace_transformation.PatternNode(dace_nodes.AccessNode)  # Needed to speed up matching.
    concat_node = dace_transformation.PatternNode(dace_nodes.AccessNode)
    map_entry = dace_transformation.PatternNode(dace_nodes.MapEntry)

    tag = dace_properties.Property(dtype=str, default=None, allow_none=True)

    # Name of all data that is used at only one place. Is computed by the
    #  `FindSingleUseData` pass and be passed at construction time. Needed until
    #  [issue#1911](https://github.com/spcl/dace/issues/1911) has been solved.
    _single_use_data: Optional[Mapping[dace.SDFG, set[str]]]

    def __init__(
        self,
        tag: Optional[str] = None,
        single_use_data: Optional[Mapping[dace.SDFG, set[str]]] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._single_use_data = single_use_data
        self.tag = tag

    @classmethod
    def expressions(cls) -> Any:
        return [
            dace.sdfg.utils.node_path_graph(
                cls.node_a1,
                cls.concat_node,
                cls.map_entry,
            )
        ]

    def can_be_applied(
        self,
        graph: Union[dace.SDFGState, dace.SDFG],
        expr_index: int,
        sdfg: dace.SDFG,
        permissive: bool = False,
    ) -> bool:
        map_entry: dace_nodes.MapEntry = self.map_entry
        concat_node: dace_nodes.AccessNode = self.concat_node

        # If we have the single use data check, perform it here to safe some scanning.
        if (
            self._single_use_data is not None
            and concat_node.data not in self._single_use_data[sdfg]
        ):
            return False

        # We can not replace global data.
        if not sdfg.arrays[concat_node.data].transient:
            return False

        # Must be on the top scope.
        if graph.scope_dict()[map_entry] is not None:
            return False

        # Check if the accesses are valid.
        if not gt_check_if_concat_where_node_is_replaceable(
            state=graph,
            concat_node=concat_node,
        ):
            return False

        # Check if single use data if not yet done.
        if self._single_use_data is None:
            find_single_use_data = dace_analysis.FindSingleUseData()
            single_use_data = find_single_use_data.apply_pass(sdfg, None)
            if concat_node.data not in single_use_data[sdfg]:
                return False

        return True

    def apply(
        self,
        graph: dace.SDFGState,
        sdfg: dace.SDFG,
    ) -> None:
        concat_node: dace_nodes.AccessNode = self.concat_node
        gt_replace_concat_where_node(
            sdfg=sdfg,
            state=graph,
            concat_node=concat_node,
            tag=self.tag,
        )


def gt_replace_concat_where_node(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    concat_node: dace_nodes.AccessNode,
    tag: Optional[str] = None,
) -> int:
    """Replaces all accesses to `concat_node`, result of a `concat_where`, with Tasklets.

    Instead of accessing the AccessNode, in which the result of the `concat_where`
    is written, directly, the function replaces accesses with Tasklets, that based
    on original access select what data from the producers are accessed.
    The function is also able to handle nested SDFGs, i.e. the `concat_where` data
    is accessed inside nested SDFGs.

    The function has currently the following limitations:
    - `concat_node` must be single use data.
    - All producers of `concat_node` must be AccessNodes (might be lifted).
    - It is only possible to replace accesses of a single element that happens on
        a Memlet, i.e. no neighborhood accesses.

    The function returns the number of edges that were replaced.

    In order to check if a `concat_where` AccessNode can be replaced use
    `gt_check_if_concat_where_node_is_replaceable()` (it is a pre condition of this
    function that `gt_check_if_concat_where_node_is_replaceable()` returned `True`).
    To replace all `concat_where` nodes in an entire SDFG
    `gt_apply_concat_where_replacement_on_sdfg()` is provided.

    Args:
        sdfg: The SDFG in which we operate.
        state: The state that contains the `concat_where` node.
        concat_node: The node that represents the result of the `concat_where` expression.
        tag: Use this to mangle the name of data descriptors that should be created.
            If not given defaults to `concat_node.data`.
    """
    assert sdfg.arrays[concat_node.data].transient

    # These are all the consumers that we need to modify.
    genuine_consumer_specs, descending_points = _find_consumer_specs_single_source_single_level(
        state, concat_node, for_check=False
    )

    scope_dict = state.scope_dict()
    initial_producer_specs = _setup_initial_producer_description_on_top_level(
        sdfg=sdfg,
        state=state,
        concat_node=concat_node,
        scope_dict=scope_dict,
    )

    # Not only the genuine consumers but also the "descending points" needs access to the data.
    producer_specs = _map_data_into_nested_scopes(
        state=state,
        scope_dict=scope_dict,
        initial_producer_specs=initial_producer_specs,
        consumer_specs=genuine_consumer_specs + descending_points,
    )

    # Now process the genuine consumers.
    nb_applies = 0
    for consumer_spec in genuine_consumer_specs:
        _replace_single_read(
            sdfg=sdfg,
            state=state,
            scope_dict=scope_dict,
            consumer_spec=consumer_spec,
            producer_specs=producer_specs,
            tag=f"{concat_node.data}_{'' if tag is None else tag}",
        )
        nb_applies += 1

    nb_applies += _process_descending_points_of_state(
        state=state,
        scope_dict=scope_dict,
        descending_points=descending_points,
        producer_specs=producer_specs,
    )

    assert state.out_degree(concat_node) == 0
    state.remove_node(concat_node)
    sdfg.remove_data(concat_node.data, validate=False)

    return nb_applies


def gt_check_if_concat_where_node_is_replaceable(
    state: dace.SDFGState,
    concat_node: dace_nodes.AccessNode,
) -> bool:
    """Check if `concat_node` is a suitable candidate for replacement by `gt_replace_concat_where_node()`.

    The function checks if the producer are valid, i.e. if they are AccessNodes and
    if the consumer of `concat_where` data can be replaced. It does not check if
    `concat_node` is single use data.

    Args:
        state: The state that contains the `concat_where` node.
        concat_node: The node that represents the result of the `concat_where` expression.
    """
    # Check the producers
    if state.in_degree(concat_node) < 2:
        # Otherwise it would be a redundant copy.
        return False
    for iedge in state.in_edges(concat_node):
        if not isinstance(iedge.src, dace_nodes.AccessNode):
            return False

    descending_points = _find_consumer_specs_single_source_single_level(
        state, concat_node, for_check=True
    )
    if descending_points is None:
        return False

    for descending_point in descending_points:
        if not _check_descending_point(descending_point):
            return False

    return True


def gt_apply_concat_where_replacement_on_sdfg(
    sdfg: dace.SDFG,
    single_use_data: Optional[Mapping[dace.SDFG, Collection[str]]] = None,
    validate: bool = False,
    validate_all: bool = False,
) -> int:
    """Applies `gt_apply_concat_where_replacement_on_sdfg()` on the entire SDFG.

    The function scans the SDFG and calls `gt_apply_concat_where_replacement_on_sdfg()`
    for all suitable nodes, including nested SDFGs. The function guarantees a stable
    ordering, that is based on the labels.
    The function returns the number of replacements.

    Args:
        sdfg: The sdfg to process.
        single_use_data: The result of the `FindSingleUseData` analysis pass. If `None`,
            the default, the function will perform the scan automatically.
        validate: Perform validation at the end.
        validate_all: Perform validation also at intermediate steps.
    """

    if single_use_data is None or sdfg not in single_use_data:
        find_single_use_data = dace_analysis.FindSingleUseData()
        single_use_data = find_single_use_data.apply_pass(sdfg, None)

    found_nsdfgs: list[tuple[dace.SDFGState, dace_nodes.NestedSDFG]] = []
    suitable_concat_nodes: list[tuple[dace.SDFGState, dace_nodes.AccessNode]] = []
    nb_applies = 0
    for state in sdfg.states():
        scope_dict = state.scope_dict()
        for node in state.nodes():
            if (
                isinstance(node, dace_nodes.AccessNode)
                and scope_dict[node] is None
                and sdfg.arrays[node.data].transient
            ):
                if node.data in single_use_data[
                    sdfg
                ] and gt_check_if_concat_where_node_is_replaceable(state, node):
                    suitable_concat_nodes.append((state, node))
            elif isinstance(node, dace_nodes.NestedSDFG):
                found_nsdfgs.append((state, node))

    if len(suitable_concat_nodes) > 0:
        suitable_concat_nodes = sorted(suitable_concat_nodes, key=lambda x: (repr(x[0]), x[1].data))
        for state, concat_node in suitable_concat_nodes:
            nb_applies += gt_replace_concat_where_node(
                sdfg=sdfg,
                state=state,
                concat_node=concat_node,
            )
            if validate_all:
                # TODO(phimuell): Limit validation to the state.
                sdfg.validate()

    if len(found_nsdfgs) > 0:
        found_nsdfgs = sorted(found_nsdfgs, key=lambda x: (repr(x[0]), str(x[1])))
        for _, nsdfg in found_nsdfgs:
            nb_applies += gt_apply_concat_where_replacement_on_sdfg(
                sdfg=nsdfg.sdfg,
                single_use_data=single_use_data,
                validate=False,
                validate_all=validate_all,
            )

    if validate:
        sdfg.validate()

    return nb_applies


_ScopeLocation: TypeAlias = Union[dace_nodes.MapEntry, None]
"""Defines a scope in the DaCe term.

This is either a `MapEntry` if nested or `None` if something is at the global scope.
"""


@dataclasses.dataclass(frozen=True)
class _DataSource:
    """Describes where (full producer) data can be found.

    Essentially a pair consisting of a Node and the required connector that must be
    used to access it.
    """

    node: Union[dace_nodes.AccessNode, dace_nodes.MapEntry]
    conn: Union[str, None]

    def __post_init__(self) -> None:
        assert isinstance(self.node, (dace_nodes.AccessNode, dace_nodes.MapEntry))


@dataclasses.dataclass(frozen=True)
class _FinalConsumerSpec:
    """Describes the final consumer."""

    edge: dace_graph.MultiConnectorEdge[dace.Memlet]

    def consumed_subset(self, state: dace.SDFGState) -> dace_sbs.Range:
        return self.edge.data.get_src_subset(self.edge, state)

    @property
    def consumer(self) -> dace_nodes.Node:
        return self.edge.dst

    def __lt__(self, other: object) -> bool:
        if isinstance(other, _FinalConsumerSpec):
            return (self.consumer, str(self.edge)) < (self.consumer, str(other.edge))
        return NotImplemented


_ConsumerPartition: TypeAlias = tuple[list[_FinalConsumerSpec], list[_FinalConsumerSpec]]
"""Represents a partition of consumer specifications.

It is a pair of `list`s. The first `list` contains the "genuine consumers", this are
the consumers that only read one element from the concat where node. Essentially,
these consumers can be handled by `_replace_single_read()`.
The second `list` contains the "descending points", these are essentially nested SDFGs,
that consume the entire data from the concat where node (a nested SDFG that just
consumes a single element is classified as a "genuine consumer"). These are the nodes
that needs further processing.
"""


@dataclasses.dataclass(frozen=True)
class _ProducerSpec:
    """Describes how a `concat_where` converges.

    This class only describes a single source of the concat where. To make sense
    one needs a `list` of `_ProducerSpec`, one for each converging edges.

    Args:
        data_name: Name of the data descriptor.
        offset: The range that is read from the data descriptor.
        subset: The range that is written into the `concat_where` node.
        desc: The data descriptor.
        data_source: Maps a scope location to a data source, i.e. the node and the
            connector where the full producer data, `self` represents, can be accessed.
            You should not access it directly but instead use `get_data_source()`.
    """

    data_name: str
    offset: dace_sbs.Range
    subset: dace_sbs.Range
    desc: dace_data.Data
    data_source: dict[_ScopeLocation, _DataSource]

    def get_data_source(
        self,
        consumer_spec: _FinalConsumerSpec,
        scope_dict: Mapping[dace_nodes.Node, _ScopeLocation],
    ) -> _DataSource:
        """Maps a consumer to the data source it should use.

        Note that `scope_dict` should be obtained by `state.scope_dict()` to prevent
        issues when the state is temporary invalid, this dict should be obtained in
        the beginning.
        """
        return self.data_source[scope_dict[consumer_spec.consumer]]

    @property
    def free_symbols(self) -> set[str]:
        # Sometimes `.free_symbols` does not returns `str` but symbols, so we have to
        #  convert them.
        return {
            str(fs)
            for fs in (self.desc.free_symbols | self.offset.free_symbols | self.subset.free_symbols)
        }

    def __copy__(self) -> "_ProducerSpec":
        return _ProducerSpec(
            data_name=self.data_name,
            offset=copy.deepcopy(self.offset),
            subset=copy.deepcopy(self.subset),
            desc=self.desc.clone(),
            data_source=self.data_source.copy(),
        )

    def __lt__(self, other: object) -> bool:
        if isinstance(other, _ProducerSpec):
            return (self.data_name, str(self.subset)) < (other.data_name, str(other.subset))
        return NotImplemented


def _setup_initial_producer_description_on_top_level(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    concat_node: dace_nodes.AccessNode,
    scope_dict: Mapping[dace_nodes.Node, _ScopeLocation],
) -> list[_ProducerSpec]:
    """Sets up the producer specifications at the top level.

    To fully populate the production specification it has to be passed to
    `_map_data_into_nested_scopes()`.

    This function should be called on the top level, i.e. not in nested SDFGs.
    For these cases you should use `_setup_initial_producer_description_in_nested_state()`.

    Args:
        sdfg: The top level SDFG, in which `state` is located.
        state: The state in which `concat_node` is located.
        concat_node: The node that is the result of a concat where expression.
        scope_dict: The scope dictionary of the state.
    """
    assert scope_dict[concat_node] is None
    # First collect the edges that define the concat where expression. We currently
    #  assume that all of them came from AccessNodes, but this is a restriction that
    #  could be lifted.
    converging_edges = list(state.in_edges(concat_node))
    assert all(isinstance(iedge.src, dace_nodes.AccessNode) for iedge in converging_edges)
    assert len(converging_edges) >= 2

    initial_producer_specs: list[_ProducerSpec] = []
    for converging_edge in converging_edges:
        converging_edge.data.try_initialize(sdfg, state, converging_edge)
        producer_node = converging_edge.src
        producer_desc = producer_node.desc(sdfg)
        offset = copy.deepcopy(converging_edge.data.src_subset)
        subset = copy.deepcopy(converging_edge.data.dst_subset)
        assert offset is not None and subset is not None

        initial_producer_specs.append(
            _ProducerSpec(
                data_name=producer_node.data,
                offset=offset,
                subset=subset,
                desc=producer_desc.clone(),
                data_source={None: _DataSource(producer_node, None)},
            )
        )
        assert all(str(fs) in sdfg.symbols for fs in initial_producer_specs[-1].free_symbols)

    return sorted(initial_producer_specs)


def _process_descending_points_of_state(
    state: dace.SDFGState,
    scope_dict: Mapping[dace_nodes.Node, _ScopeLocation],
    descending_points: Sequence[_FinalConsumerSpec],
    producer_specs: Sequence[_ProducerSpec],
) -> int:
    """Processes all descending points inside a single state.

    Processes all descending points in the given state. It will configure them if
    needed. However, the function assumes that the producer data has already been
    made available inside the scopes of the descending points, which is the case
    if `_map_data_into_nested_scopes()` has been run on the state.

    Args:
        state: The state containing the descending points.
        scope_dict: The scope dict of `state`.
        descending_points: All descending points inside `state`.
        producer_specs: The producer specification of the parent.
    """

    configured_descending_points: dict[dace_nodes.NestedSDFG, list[_ProducerSpec]] = {}
    nb_applies = 0
    for descending_point in descending_points:
        # If needed configure the descending point.
        nsdfg: dace_nodes.NestedSDFG = descending_point.consumer
        if nsdfg not in configured_descending_points:
            configured_descending_points[nsdfg] = _configure_descending_point(
                state=state,
                scope_dict=scope_dict,
                descending_point=descending_point,
                parent_producer_specs=producer_specs,
            )

        # Process this descending point.
        # NOTE: A nested SDFG might be processed multiple times, but each time
        #   different data is processed.
        nb_applies += _process_descending_points_of_state_impl(
            this_descending_point=descending_point,
            nested_initial_producer_specs=configured_descending_points[nsdfg],
        )

        # Now remove the Memlet "path" that mapped the concat where data into the
        #  nested SDFG and also delete its alias inside it.
        _cleanup_memlet_path(state, descending_point)
        descending_point.consumer.remove_in_connector(descending_point.edge.dst_conn)
        nsdfg.sdfg.remove_data(descending_point.edge.dst_conn, validate=gtx_config.DEBUG)

    return nb_applies


def _process_descending_points_of_state_impl(
    this_descending_point: _FinalConsumerSpec,
    nested_initial_producer_specs: list[_ProducerSpec],
) -> int:
    """Process the interior of a descending point.

    Essentially iterate through all states of the descending point and perform
    the replacement of concat where accesses. The function assumes that
    `this_descending_point` has been configured.
    Furthermore, the function will handled descending points that are located within
    `this_descending_point` by passing them to `_process_descending_points_of_state()`.

    Args:
        this_descending_point: The descending point that is currently processed.
        nested_initial_producer_specs: The producer specification that describes
            the concat where convergence in the context of this descending point.
            It is obtained by a call to `_configure_descending_point()`.

    Note:
        - This function should not be called directly.
        - It is possible that a nested SDFG is handled multiple times, but each time
            it is handled through a different descending point, i.e. the concat where
            data is aliased to multiple data.
    """
    assert all(len(pspec.data_source) == 0 for pspec in nested_initial_producer_specs)

    consumers = _find_consumer_specs_in_descending_point(this_descending_point)
    nb_applies = 0
    for state, (_concat_node, (genuine_consumer_specs, descending_points)) in consumers.items():
        initial_producer_specs = _setup_initial_producer_description_in_nested_state(
            descending_point=this_descending_point,
            nested_state=state,
            nested_initial_producer_specs=nested_initial_producer_specs,
        )
        # `state.scope_dict()` must be called after `_setup_initial_producer_description_in_nested_state()`.
        scope_dict = state.scope_dict()
        producer_specs = _map_data_into_nested_scopes(
            state=state,
            scope_dict=scope_dict,
            initial_producer_specs=initial_producer_specs,
            consumer_specs=genuine_consumer_specs + descending_points,
        )

        # Replace the consumer on this level.
        for consumer_spec in genuine_consumer_specs:
            _replace_single_read(
                sdfg=state.sdfg,
                state=state,
                scope_dict=scope_dict,
                consumer_spec=consumer_spec,
                producer_specs=producer_specs,
                tag=f"{this_descending_point.edge.dst_conn}",
            )
            nb_applies += 1

        # Now recursively descend.
        if len(descending_points) > 0:
            nb_applies += _process_descending_points_of_state(
                state=state,
                scope_dict=scope_dict,
                descending_points=descending_points,
                producer_specs=producer_specs,
            )

        # NOTE: `_concat_node` has been removed already by `_configure_descending_point()`.

    # NOTE: We can not remove the concat where data here, because a descending point
    #   does not equals a nested SDFG. Thus there could be other descending points
    #   that refers to the same nested SDFG. The data is removed in
    #   `_process_descending_points_of_state()`.

    return nb_applies


def _setup_initial_producer_description_in_nested_state(
    descending_point: _FinalConsumerSpec,
    nested_state: dace.SDFGState,
    nested_initial_producer_specs: Sequence[_ProducerSpec],
) -> list[_ProducerSpec]:
    """Analogue operation to `_setup_initial_producer_description_on_top_level()` but on nested levels.

    The function expects that `initial_producer_specs` was generated by a previous
    call to `_configure_descending_point()`, that got the same `descending_point`.
    Furthermore, the function expects that `nested_state` is a state inside the
    descending point.

    In principle the function will do the same as `_setup_initial_producer_description_on_top_level()`,
    i.e. setting up the producer and scope source description. However, it will also
    generate AccessNodes for the mapped data in the state if they are not yet there.

    Args:
        descending_point: The descending point in which we operate.
        nested_state: A state inside the nested SDFG.
        nested_initial_producer_specs: The transformed initial producer
            specification, generated by `_configure_descending_point()`.
    """
    nested_sdfg: dace.SDFG = descending_point.consumer.sdfg
    assert nested_state in nested_sdfg.states()
    assert all(len(prod_spec.data_source) == 0 for prod_spec in nested_initial_producer_specs)

    # Look for access nodes that already exist.
    source_access_nodes: dict[str, dace_nodes.AccessNode] = {}
    needed_data: set[str] = {
        producer_spec.data_name for producer_spec in nested_initial_producer_specs
    }
    for dnode in nested_state.data_nodes():
        if dnode.data in needed_data:
            needed_data.remove(dnode.data)
            assert not nested_sdfg.arrays[dnode.data].transient
            assert nested_state.scope_dict()[dnode] is None
            source_access_nodes[dnode.data] = dnode
            if len(needed_data) == 0:
                break

    # Now create the ones that are missing.
    for missing_data in needed_data:
        assert not nested_sdfg.arrays[missing_data].transient
        source_access_nodes[missing_data] = nested_state.add_access(missing_data)

    return [
        _ProducerSpec(
            data_name=prod_spec.data_name,
            offset=copy.deepcopy(prod_spec.offset),
            subset=copy.deepcopy(prod_spec.subset),
            desc=prod_spec.desc.clone(),
            data_source={None: _DataSource(source_access_nodes[prod_spec.data_name], None)},
        )
        for prod_spec in nested_initial_producer_specs
    ]


def _configure_descending_point(
    state: dace.SDFGState,
    scope_dict: Mapping[dace_nodes.Node, _ScopeLocation],
    descending_point: _FinalConsumerSpec,
    parent_producer_specs: Sequence[_ProducerSpec],
) -> list[_ProducerSpec]:
    """Handles a descending point.

    The function computes the initial producer specification for the nested SDFG.
    In addition it ensures:
    - That the nested SDFG has the necessary data descriptors and symbols:
    - That the nested SDFG has an updated symbol mapping.
    - That the nested SDFG was connected to the data source.
    - Applies the symbol remapping to the returned producer specifications.

    The function returns a transformed version of the initial producer specification
    that can be used inside the descending point. It must later be passed to
    `_setup_initial_producer_description_in_nested_state()` to obtain a full
    producer specification for a particular state inside the nested SDFG.

    It is important that this function can only be called once per descending point.
    Furthermore, it will not remove the connection between the concat where node
    and the nested SDFG, i.e. `descending_point.edge`.

    Args:
        state: The state containing the descending point.
        scope_dict: The scope dict of `state`.
        descending_point: The consumer that represents a descending point, located in `scope`.
        parent_producer_specs: The initial producer specification on the top level.

    Note:
        You should not use this function directly, instead use `_process_descending_points_of_state()`.
    """
    assert isinstance(descending_point.consumer, dace_nodes.NestedSDFG)
    nsdfg_node: dace_nodes.NestedSDFG = descending_point.consumer
    nsdfg: dace.SDFG = nsdfg_node.sdfg
    symbol_mapping: dict[str, dace_sym.SymExpr] = nsdfg_node.symbol_mapping

    # Now we need to map the data into the nested SDFG. The mapping function only
    #  gives us `inter -> top` so we have to invert it. Inversion here should be
    #  stable because the data is internally ordered.
    nested_to_parent_name_mapping = gtx_transformations.utils.gt_data_descriptor_mapping(
        state=state,
        nsdfg=nsdfg_node,
        only_fully_mapped=True,
        only_inputs=True,
    )
    parent_to_nested_name_mapping = {
        parent: nested for nested, parent in nested_to_parent_name_mapping.items()
    }

    nested_initial_producer_specs: list[_ProducerSpec] = []
    handled_data: dict[str, _ProducerSpec] = {}
    full_repl_dict: dict[str, str] = {}  # All known replacements.
    for parent_prod_spec in parent_producer_specs:
        parent_data_name = parent_prod_spec.data_name

        # Compute the replacement mapping for symbols.
        repl_dict: dict[str, str] = {}  # For this producer.
        for psym in sorted(parent_prod_spec.free_symbols):
            ptype = state.sdfg.symbols[psym]
            if psym in full_repl_dict:
                # The symbol was already handled in a previous producer, so we reuse it.
                nested_symbol = full_repl_dict[psym]
                assert nested_symbol in nsdfg.symbols and nsdfg.symbols[nested_symbol] == ptype

            elif (
                psym in symbol_mapping
                and psym == str(symbol_mapping[psym])
                and psym in nsdfg.symbols
                and nsdfg.symbols[psym] == ptype
            ):
                # The symbol is already mapped 1:1 into the nested sdfg -> nothing to do.
                nested_symbol = psym
                assert nested_symbol not in repl_dict

            else:
                # The symbol is either not known or it is not mapped 1:1 or has the wrong type,
                #  in either case we have to create a new symbol inside the nested SDFG.
                nested_symbol = nsdfg.add_symbol(psym, ptype, find_new_name=True)
                assert nested_symbol not in symbol_mapping
                symbol_mapping[nested_symbol] = psym

            if psym != nested_symbol:
                repl_dict[psym] = nested_symbol
                if psym not in full_repl_dict:
                    full_repl_dict[psym] = nested_symbol

        # Compute the nested data descriptor.
        if parent_data_name in handled_data:
            # The data has already been handled before. So we can simply reuse it.
            #  Symbol renaming has already been handled.
            already_handled_prod_spec = handled_data[parent_data_name]
            nested_desc = already_handled_prod_spec.desc.clone()

        elif parent_data_name in parent_to_nested_name_mapping:
            # We have not handled the data, but it is already mapped into the nested
            #  SDFG, for other reasons. Thus we will reuse this data descriptor.
            #  The symbol renaming is already done and not necessarily described
            #  through `repl_dict`.
            nested_data_name = parent_to_nested_name_mapping[parent_data_name]
            nested_desc = nsdfg.arrays[nested_data_name].clone()
            assert not nested_desc.transient

        else:
            # The data is not yet mapped into the nested SDFG, so we have to create
            #  a new one. And we have to apply symbol renaming on it. In addition
            #  we have to make sure that it is a global.
            nested_desc = parent_prod_spec.desc.clone()
            nested_desc.transient = False
            if repl_dict:
                dace_sym.safe_replace(
                    mapping=repl_dict,
                    replace_callback=functools.partial(
                        dace.sdfg.replace_properties_dict, nested_desc
                    ),
                )
            nested_data_name = nsdfg.add_datadesc(parent_data_name, nested_desc, find_new_name=True)

            # We also need to map it into the nested SDFG.
            source_loc = parent_prod_spec.get_data_source(descending_point, scope_dict)
            state.add_edge(
                source_loc.node,
                source_loc.conn,
                descending_point.consumer,
                nested_data_name,
                dace.Memlet(
                    data=parent_prod_spec.data_name,
                    subset=dace_sbs.Range.from_array(parent_prod_spec.desc),
                ),
            )
            assert nested_data_name not in descending_point.consumer.in_connectors
            descending_point.consumer.add_in_connector(
                nested_data_name, dtype=dace.pointer(nested_desc.dtype)
            )

        # Apply symbol renaming on offset and subset; descriptor was handled above.
        nested_offset = copy.deepcopy(parent_prod_spec.offset)
        nested_subset = copy.deepcopy(parent_prod_spec.subset)
        if repl_dict:
            for nested_set in [nested_offset, nested_subset]:
                dace_sym.safe_replace(  # Performs inplace replacement
                    mapping=repl_dict,
                    replace_callback=nested_set.replace,
                )

        nested_initial_producer_specs.append(
            _ProducerSpec(
                data_name=nested_data_name,
                offset=nested_offset,
                subset=nested_subset,
                desc=nested_desc,
                data_source={},  # Intentionally empty.
            )
        )

    # Restore the proper format of the symbol mapping.
    nsdfg_node.symbol_mapping = symbol_mapping

    return nested_initial_producer_specs


def _map_data_into_nested_scopes(
    state: dace.SDFGState,
    scope_dict: Mapping[dace_nodes.Node, _ScopeLocation],
    initial_producer_specs: Sequence[_ProducerSpec],
    consumer_specs: Sequence[_FinalConsumerSpec],
) -> list[_ProducerSpec]:
    """Complete the scope information and ensure that they are available.

    The function returns a finalized version of the production specifications passed
    through `initial_producer_specs`, which were obtained by a previous call to
    `_setup_initial_producer_description_on_top_level()` or
    `_setup_initial_producer_description_in_nested_state()`.
    The main differences between the return value and `initial_producer_specs` is
    that `_ProducerSpec::data_source` has been fully populated and the scope nodes
    are prepared.

    Args:
        state: The state in which we operate.
        scope_dict: The scope dict of `state`.
        initial_producer_specs: The initial producers.
        consumer_specs: All consumers that should be prepared.

    Note:
        In most cases you should pass both the genuine consumer and the descending
        points to this function.
    """
    # Pre-Condition: The top level AccessNodes, which constitute the producers of the
    #  `concat_where` expression must be handled already.
    assert all(None in prod_spec.data_source for prod_spec in initial_producer_specs)
    producer_specs = [copy.copy(producer_spec) for producer_spec in initial_producer_specs]

    # These are the scopes in which _accesses_ to the `concat_where` happens, It does
    #  not (necessarily) contains their parent scopes as well. Note that in order to
    #  make the data accessible in a scope, the data needs to be accessible in all of
    #  its ancestors scopes too. We will bring them in a deterministic order before
    #  process them, handling missing parent scopes on the fly, see
    #  `_map_data_into_nested_scopes_impl()`.
    scopes_containing_consumers: list[_ScopeLocation] = sorted(
        {scope_dict[consumer_spec.consumer] for consumer_spec in consumer_specs},
        key=lambda scope: "NONE" if scope is None else str(scope),
    )
    for scope in scopes_containing_consumers:
        _map_data_into_nested_scopes_impl(
            state=state,
            scope_dict=scope_dict,
            scope_to_handle=scope,
            producer_specs=producer_specs,
        )

    return producer_specs


def _map_data_into_nested_scopes_impl(
    state: dace.SDFGState,
    scope_dict: Mapping[dace_nodes.Node, _ScopeLocation],
    scope_to_handle: _ScopeLocation,
    producer_specs: Sequence[_ProducerSpec],
) -> None:
    """Helper function of `_map_data_into_nested_scopes()`.

    This function will ensure that the producers of the `concat_where` expression are
    accessible inside `scope_to_handle`. The information is returned inside the
    `data_source` attribute of the passed `producer_specs` objects.
    This is either done by reusing an already existing suitable Memlet or creating
    a new one.

    If the data is not accessible inside the parent scope of `scope_to_handle`, all
    missing ancestor scopes will be handled first.
    Furthermore, it is not an error if `scope_to_handle` was already handled. This
    might happen in case `scope_to_handle` was the parent of a scope previously
    handled (think of nested scopes each containing proper consumers and they are
    processing starting from the most deeply nested one).

    The only condition this function has is that the top level, i.e. `scope_to_handle`
    is `None`, has already been handled.
    """

    # Test if the scope is already handled. It is enough to check the first producer
    #  because if a scope was handled for one producer it was handled for all.
    if scope_to_handle in producer_specs[0].data_source:
        assert all(scope_to_handle in producer_spec.data_source for producer_spec in producer_specs)
        return

    # The global scope, i.e. the AccessNodes referring to the data, has already been
    #  handled outside, thus `scope_to_handle` (at this point) can only be a `MapEntry`.
    assert isinstance(scope_to_handle, dace_nodes.EntryNode)

    # Check if we have to handle the parent scope first. Again it is enough to check
    #  the first producer.
    parent_scope = scope_dict[scope_to_handle]
    if parent_scope not in producer_specs[0].data_source:
        _map_data_into_nested_scopes_impl(
            state=state,
            scope_dict=scope_dict,
            scope_to_handle=parent_scope,
            producer_specs=producer_specs,
        )

    # Now connect the data from `parent_scope` to `scope_to_handle`.
    already_mapped_data: dict[str, _DataSource] = {}
    for producer_spec in producer_specs:
        assert parent_scope in producer_spec.data_source
        assert scope_to_handle not in producer_spec.data_source

        if producer_spec.data_name in already_mapped_data:
            # This data was already seen so reuse the connector, no need to create
            #  an additional edge.
            data_source = already_mapped_data[producer_spec.data_name]

        else:
            # This data has never been seen. Check if there is a already a connection
            #  between scope and its parent (for some other reason). We have to make
            #  sure that the full data is fully transferred.
            parent_source = producer_spec.data_source[parent_scope]
            full_shape = dace_sbs.Range.from_array(producer_spec.desc)  # Might be consumed later.
            potential_reusable_edge = next(
                (
                    oedge
                    for oedge in state.out_edges_by_connector(
                        parent_source.node, parent_source.conn
                    )
                    if (
                        (not oedge.data.is_empty())
                        and oedge.dst is scope_to_handle
                        and oedge.dst_conn.startswith("IN_")
                        # For syntactical reasons one has to use `src_subset` here.
                        and oedge.data.get_src_subset(oedge, state).covers(full_shape)
                    )
                ),
                None,
            )

            if potential_reusable_edge is not None:
                # There was an edge that we can reuse.
                scope_connector = "OUT_" + potential_reusable_edge.dst_conn[3:]
            else:
                # Create a new edge between the `parent_scope` and `scope_to_handle` nodes.
                new_conn_name = scope_to_handle.next_connector(producer_spec.data_name)
                state.add_edge(
                    parent_source.node,
                    parent_source.conn,
                    scope_to_handle,
                    "IN_" + new_conn_name,
                    dace.Memlet(
                        data=producer_spec.data_name,
                        subset=full_shape,  # Intentionally not copied.
                    ),
                )
                scope_to_handle.add_scope_connectors(new_conn_name)
                scope_connector = f"OUT_{new_conn_name}"
            data_source = _DataSource(scope_to_handle, scope_connector)

        producer_spec.data_source[scope_to_handle] = data_source
        if producer_spec.data_name not in already_mapped_data:
            already_mapped_data[producer_spec.data_name] = data_source


def _replace_single_read(
    sdfg: dace.SDFG,
    state: dace.SDFGState,
    scope_dict: Mapping[dace_nodes.Node, _ScopeLocation],
    consumer_spec: _FinalConsumerSpec,
    producer_specs: Sequence[_ProducerSpec],
    tag: str,
) -> None:
    """Performs the replacement of a single read, defined by `consumer_spec`.

    Essentially the function will replace the access to a converging `concat_where`
    node, i.e. the node where multiple writes converges, with a Tasklet that selects
    the data that should be read. This means only the read of a single element can
    be handled in this way.

    Note that the final consumer node will remain in the SDFG, instead the function
    will create a new data descriptor, into which the result will be written that
    will then be connected to the final consumer. However, the edge will be removed.

    Args:
        sdfg: The sdfg that we process.
        state: The state on which we operate.
        scope_dict: The scope dict of `state`.
        consumer_spec: Describes a consumer.
        producer_specs: Descriptions of how the `concat_where` converges.
        tag: Used to give newly created elements a unique name.
    """
    assert consumer_spec.edge.data.wcr is None
    assert consumer_spec.edge.data.subset.num_elements() == 1

    # For every producer specification one entry. The names might not be unique, in
    #  case two producers refer to the same data the corresponding input names will
    #  be the same.
    tlet_inputs: list[str] = []
    tlet_output = "__out"

    # Maps the name of a producer (`_ProducerSpec::data_name`) to a pair. The first
    #  element is the corresponding Tasklet input connector and second is _a_
    #  `_ProducerSpec`. Note in this context the `offset` and `subset` attribute
    #  have no meaning.
    tinput_map: dict[str, tuple[str, _ProducerSpec]] = {}

    select_conds: list[str] = []
    prod_accesses: list[str] = []
    consumed_subset = consumer_spec.consumed_subset(state)
    for i in range(len(producer_specs)):
        producer_spec = producer_specs[i]
        prod_subset = producer_spec.subset
        prod_offsets = producer_spec.offset

        this_select_cond: list[str] = []
        this_prod_access: list[str] = []
        for dim in range(consumed_subset.dims()):
            consumer_access = consumed_subset[dim][0]
            prod_supply_start = prod_subset[dim][0]
            prod_supply_end = prod_subset[dim][1]  # Inclusive end.
            prod_offset = prod_offsets[dim][0]

            this_select_cond.append(
                f"((({prod_supply_start}) <= ({consumer_access})) and (({consumer_access}) <= ({prod_supply_end})))"
            )
            this_prod_access.append(
                f"(({prod_offset}) + (({consumer_access}) - ({prod_supply_start})))"
            )
        prod_accesses.append(", ".join(this_prod_access))
        select_conds.append(" and ".join(this_select_cond))

        # Generate the name of the input that is used.
        if producer_spec.data_name in tinput_map:
            # Data is known reuse it.
            tlet_inputs.append(tinput_map[producer_spec.data_name][0])
        else:
            # Data is not known, so create a new name.
            tlet_inputs.append(f"__in{i}")
            tinput_map[producer_spec.data_name] = (tlet_inputs[-1], producer_spec)

    # This writes the Tasklet code as a series of nested `?:` operators.
    def write_tasklet_code(tlet_inputs, select_conds, prod_accesses):  # type: ignore[no-untyped-def]
        if len(tlet_inputs) == 2:
            return f"{tlet_inputs[0]}[{prod_accesses[0]}] if ({select_conds[0]}) else {tlet_inputs[1]}[{prod_accesses[1]}]"
        assert len(tlet_inputs) > 2
        return f"{tlet_inputs[0]}[{prod_accesses[0]}] if ({select_conds[0]}) else ({write_tasklet_code(tlet_inputs[1:], select_conds[1:], prod_accesses[1:])})"

    tlet_code = f"{tlet_output} = {write_tasklet_code(tlet_inputs, select_conds, prod_accesses)}"

    # Create the access Tasklet.
    names_of_existing_tasklets = {
        node.label for node in state.nodes() if isinstance(node, dace_nodes.Tasklet)
    }
    tasklet_name = dace.utils.find_new_name(
        f"concat_where_tasklet_{tag}", names_of_existing_tasklets
    )
    concat_where_tasklet = state.add_tasklet(
        tasklet_name,
        inputs=set(tlet_inputs),
        outputs={tlet_output},
        code=tlet_code,
    )

    # Connect the tasklet to the source.
    for data_name, (tlet_input, producer_spec) in tinput_map.items():
        assert data_name == producer_spec.data_name
        data_source = producer_spec.get_data_source(consumer_spec, scope_dict)
        state.add_edge(
            data_source.node,
            data_source.conn,
            concat_where_tasklet,
            tlet_input,
            dace.Memlet(
                data=data_name,
                subset=dace_sbs.Range.from_array(producer_spec.desc),
            ),
        )

    # Instead of replacing `final_consumer` we create an intermediate output node.
    #  This removes the need to reconfigure the dataflow.
    intermediate_data, _ = sdfg.add_scalar(
        name=f"__gt4py_concat_where_mapper_temp_{tag}",
        dtype=sdfg.arrays[consumer_spec.edge.data.data].dtype,
        transient=True,
        find_new_name=True,
    )
    intermediate_node = state.add_access(intermediate_data)
    state.add_edge(
        concat_where_tasklet,
        tlet_output,
        intermediate_node,
        None,
        dace.Memlet(data=intermediate_data, subset="0"),
    )

    # Find out what `other_subset` of the new Memlet should be.
    if (
        isinstance(consumer_spec.consumer, dace_nodes.AccessNode)
        and consumer_spec.consumer.data == consumer_spec.edge.data.data
    ):
        other_subset = consumer_spec.edge.data.subset
    else:
        other_subset = consumer_spec.edge.data.other_subset

    # Create the edge between the new intermediate and the old consumer. We ignore
    #  some properties of the Memlet here, such as dynamic (`wcr` was handled above).
    state.add_edge(
        intermediate_node,
        None,
        consumer_spec.consumer,
        consumer_spec.edge.dst_conn,
        dace.Memlet(
            data=intermediate_data,
            subset="0",
            other_subset=other_subset,
        ),
    )

    # Now remove the old Memlet path.
    _cleanup_memlet_path(state, consumer_spec)


def _find_consumer_specs_in_descending_point(
    descending_point: _FinalConsumerSpec,
) -> dict[dace.SDFGState, tuple[dace_nodes.AccessNode, _ConsumerPartition]]:
    """Scans the descending point for consumer (non recursive).

    The function will scan every state for occurrences of the concat where data
    and return the consumer partition for each state. If there are multiple
    AccessNode referring to the concat where data then the nodes are unified,
    i.e. this function mutates the state.

    Furthermore, this function assumes that the partition exists, i.e.
    `_check_descending_point()` has returned `True`.
    """
    sdfg: dace.SDFG = descending_point.consumer.sdfg
    concat_data = descending_point.edge.dst_conn

    consumers: dict[dace.SDFG, tuple[dace_nodes.AccessNode, _ConsumerPartition]] = {}
    for state in sdfg.states():
        concat_node: Union[None, dace_nodes.AccessNode] = None
        for dnode in state.data_nodes():
            if dnode.data != concat_data:
                continue

            assert state.in_degree(dnode) == 0
            if concat_node is None:
                concat_node = dnode
            else:
                # We found a second node referring to the concat where data,
                #  we have to merge them.
                for oedge in state.out_edges(dnode):
                    dace_transformation.helpers.redirect_edge(
                        state=state, edge=oedge, new_src=concat_node
                    )
                assert state.degree(dnode) == 0
                state.remove_node(dnode)

        if concat_node is None:
            continue

        consumer_partition = _find_consumer_specs_single_source_single_level(
            state, concat_node, for_check=False
        )
        consumers[state] = (concat_node, consumer_partition)
    return consumers


def _check_descending_point(
    descending_point: _FinalConsumerSpec,
) -> bool:
    """Check if this descending point has valid consumers, consumer are checked recursively."""
    sdfg: dace.SDFG = descending_point.consumer.sdfg
    concat_data = descending_point.edge.dst_conn
    assert not sdfg.arrays[concat_data].transient

    for state in sdfg.states():
        for dnode in state.data_nodes():
            if dnode.data != concat_data:
                continue
            if state.in_degree(dnode) != 0:
                return False
            scan_for_nested_descending_points = _find_consumer_specs_single_source_single_level(
                state, dnode, for_check=True
            )
            if scan_for_nested_descending_points is None:
                return False
            for nested_descending_point in scan_for_nested_descending_points:
                if not _check_descending_point(nested_descending_point):
                    return False
    return True


@overload
def _find_consumer_specs_single_source_single_level(
    state: dace.SDFGState,
    concat_node: dace_nodes.AccessNode,
    for_check: Literal[False],
) -> _ConsumerPartition: ...


@overload
def _find_consumer_specs_single_source_single_level(
    state: dace.SDFGState,
    concat_node: dace_nodes.AccessNode,
    for_check: Literal[True],
) -> Optional[list[_FinalConsumerSpec]]: ...


def _find_consumer_specs_single_source_single_level(
    state: dace.SDFGState,
    concat_node: dace_nodes.AccessNode,
    for_check: bool,
) -> Union[Optional[list[_FinalConsumerSpec]], _ConsumerPartition]:
    """Find all consumers of `concat_node` in state `state`.

    If `for_check` is `False` then the function returns the consumer partition for
    `concat_node`. Note if the consumer partition does not exists an error is generated.
    If `for_check` is `True` then the function only checks if the partition exists,
    which is indicated by a list containing all descending points of this state,
    can be empty. `None` is returned if the partition does not exist.
    To further check the returned descending points you can use `_check_descending_point()`.

    Important this function does not:
    - Considering any other state.
    - Considering any other AccessNode in `state` that also refers to the same
        data as `concat_node` does.
    - Does not descend into nested SDFGs.

    Args:
        state: The state in which we operate.
        concat_node: The node representing the `concat_where` result.
        for_check: Only perform checks.

    Note:
        If `for_check` is `False` then both kind of consumers are sorted. For `True`
        the order of the returned descending points is unspecific.
    """
    if state.out_degree(concat_node) == 0:
        return ([], []) if not for_check else []

    # First collect all consumer, we will check and partition them afterwards.
    all_consumers: list[_FinalConsumerSpec] = []
    edges_to_scan_next: list[dace_graph.MultiConnectorEdge[dace.Memlet]] = list(
        state.out_edges(concat_node)
    )
    partition_does_not_exists = False
    while len(edges_to_scan_next) != 0:
        edges_to_scan, edges_to_scan_next = edges_to_scan_next, []
        for oedge in edges_to_scan:
            consumer = oedge.dst
            if oedge.data.is_empty():
                partition_does_not_exists = True  # We should probably handle it.
                break
            if isinstance(consumer, dace_nodes.MapEntry):
                if not oedge.dst_conn.startswith("IN_"):
                    partition_does_not_exists = True  # Dynamic map ranges are not supported.
                    break
                for edge_inside_map in state.out_edges_by_connector(
                    consumer, "OUT_" + oedge.dst_conn[3:]
                ):
                    edges_to_scan_next.extend(
                        edge for edge in state.memlet_tree(edge_inside_map).leaves()
                    )
            elif isinstance(
                consumer, (dace_nodes.AccessNode, dace_nodes.Tasklet, dace_nodes.NestedSDFG)
            ):
                all_consumers.append(_FinalConsumerSpec(edge=oedge))
            else:
                partition_does_not_exists = True  # These kind of consumers can not be handled.
                break
        if partition_does_not_exists:
            if for_check:
                return None
            raise ValueError(f"Expected that partition for {concat_node} exists but it does not.")

    # Now wet and partition the consumer. Most importantly test if they only read
    #  a single element, in which case they are genuine consumers this applies to
    #  _all_ nodes, even nested SDFGs. Otherwise it might open up a "new level".
    #  For this the consumer must be a nested SDFG that reads the _whole_ concat
    #  where node.
    concat_where_shape = dace_sbs.Range.from_array(concat_node.desc(state.sdfg))
    stairs_to_deeper_levels: list[_FinalConsumerSpec] = []  # Better name appreciated.
    consumer_specs: list[_FinalConsumerSpec] = []
    for consumer_spec in all_consumers:
        consumed_subset = consumer_spec.consumed_subset(state)
        if consumed_subset.num_elements() == 1:  # Real consumer
            if not for_check:
                consumer_specs.append(consumer_spec)
        elif isinstance(consumer_spec.consumer, dace_nodes.NestedSDFG) and (
            consumed_subset.covers(concat_where_shape)
            or _handle_special_case_of_gt4py_scan_point(
                state, consumer_spec, concat_node, consumed_subset
            )
        ):
            # The whole array is mapped into the nested SDFG.
            stairs_to_deeper_levels.append(consumer_spec)
        else:
            if for_check:
                return None
            raise ValueError(f"Expected that partition for {concat_node} exists but it does not.")

    if for_check:
        return stairs_to_deeper_levels
    return sorted(consumer_specs), sorted(stairs_to_deeper_levels)


def _handle_special_case_of_gt4py_scan_point(
    state: dace.SDFGState,
    descending_point: _FinalConsumerSpec,
    concat_node: dace_nodes.AccessNode,
    consumed_subset: dace_sbs.Range,
) -> bool:
    """Performs special checking for cases where data is not properly mapped into a nested SDFG.

    In certain cases, especially for scans, DaCe Memlet propagation (or related) will
    fail to properly annotate the consumer Memlet that maps the data inside a nested
    SDFG. This function applies some heuristic checks if the nested SDFG can be
    processed or not.

    The pathological case is a nested SDFG inside a Map scope, which is used for the
    scan expressions. By construction, the scan nested SDFG does access to a single
    point in the horizontal domain. However, for unknown reason, DaCe replaces the
    point access with a subset, as an example `[i_Cell_gtx_horizontal, 0:3]` becomes
    `[0:i_Cell_gtx_horizontal + 1, 0:3]`. Thus, `consumed_subset` is not a point
    nor the full shape either, but we know that it is safe to apply the transformation.
    """
    assert isinstance(descending_point.consumer, dace_nodes.NestedSDFG)
    nsdfg: dace_nodes.NestedSDFG = descending_point.consumer

    if _handle_special_case_of_gt4py_scan_point_impl(
        state, descending_point, concat_node, consumed_subset
    ):
        warnings.warn(
            f"Special rule applied to `concat_where`-inline `{concat_node.data}` into `{nsdfg.label}`.",
            stacklevel=1,
        )
        return True
    else:
        warnings.warn(
            f"Special rule applied to `concat_where`-inline `{concat_node.data}` into `{nsdfg.label}` was rejected.",
            stacklevel=1,
        )
        return False


def _handle_special_case_of_gt4py_scan_point_impl(
    state: dace.SDFGState,
    descending_point: _FinalConsumerSpec,
    concat_node: dace_nodes.AccessNode,
    consumed_subset: dace_sbs.Range,
) -> bool:
    # There must be one node in the nested SDFG that is not a state.
    # TODO(phimuell): Come up with a better test.
    nsdfg: dace_nodes.NestedSDFG = descending_point.consumer
    if all(isinstance(node, dace.SDFGState) for node in nsdfg.sdfg.nodes()):
        return False

    # At this point `state` should be in a valid state so that it is safe to
    #  compute the scope dict.
    scope_dict = state.scope_dict()
    if scope_dict[nsdfg] is None:
        return False

    parent_desc = concat_node.desc(state.sdfg)
    nested_desc = nsdfg.sdfg.arrays[descending_point.edge.dst_conn]

    if type(parent_desc) is not type(nested_desc):
        return False
    # TODO(phimuell): Make the test involving the shape a bit smarter.
    if len(parent_desc.shape) != len(nested_desc.shape):
        return False

    if not all((start == 0) == True for start in consumed_subset.min_element()):  # noqa: E712 [true-false-comparison]  # SymPy comparison
        return False

    # Find all map parameters.
    edge = descending_point.edge
    map_params: set[str] = set()
    while isinstance(edge.src, dace_nodes.MapEntry):
        assert edge.src_conn.startswith("OUT")
        map_params.update(edge.src.map.params)
        edge = next(
            iter(e for e in state.in_edges_by_connector(edge.src, "IN_" + edge.src_conn[4:]))
        )

    assert scope_dict[edge.dst] is None  # really strange case.

    for i, end in enumerate(consumed_subset.max_element()):
        end_str = str(end)
        if end_str.isdigit():
            if (int(end_str) + 1) != parent_desc.shape[i]:  # `+1` because of storage format.
                return False
        else:
            # It is a symbol, we only check for the pathological case, i.e. a Map
            #  parameter is the end value. We also allow only one value.
            if end_str not in map_params:
                return False
            else:
                map_params.discard(end_str)

    return True


def _cleanup_memlet_path(state: dace.SDFGState, consumer_spec: _FinalConsumerSpec) -> None:
    """Similar to `remove_edge_and_dangling_path()` but special to our case.

    Removes the full path that leads to `consumer_spec` and removing the ultimate
    source if it has become empty. The connector name of the original consumer
    will never be removed.
    """

    edge: dace_graph.MultiConnectorEdge[dace.Memlet] = consumer_spec.edge
    while True:
        state.remove_edge(edge)

        # We go up as long as `edge.src` is a scoping node.
        if not isinstance(edge.src, dace_nodes.EntryNode):
            break

        # If `edge.src_conn` is in use we are done, otherwise we remove it.
        if len(list(state.out_edges_by_connector(edge.src, edge.src_conn))) != 0:
            break

        # We now find the next edge, which must be done before we remove the connectors.
        other_conn = "IN_" + edge.src_conn[4:]
        next_edges = list(state.in_edges_by_connector(edge.src, other_conn))
        assert len(next_edges) == 1

        # Now remove the connectors.
        edge.src.remove_out_connector(edge.src_conn)
        edge.src.remove_in_connector(other_conn)

        edge = next_edges[0]

    # Test for isolated node and remove it in that case
    if (state.in_degree(edge.src) == 0) and (state.out_degree(edge.src) == 0):
        state.remove_node(edge.src)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

"""
Interface for program processors.

Program processors are functions which operate on a program paired with the input
arguments for the program. Programs are represented by an ``iterator.ir.Program``
node. Program processors that execute the program with the given arguments (possibly by generating
code along the way) are program executors. Those that generate any kind of string based
on the program and (optionally) input values are program formatters.

For more information refer to
``gt4py/docs/functional/architecture/007-Program-Processors.md``
"""

from __future__ import annotations

import abc
import dataclasses
from typing import Any, Callable

import gt4py.next.iterator.ir as itir


class ProgramFormatter:
    @abc.abstractmethod
    def __call__(self, program: itir.Program, *args: Any, **kwargs: Any) -> str: ...


@dataclasses.dataclass(frozen=True)
class WrappedProgramFormatter(ProgramFormatter):
    formatter: Callable[..., str]

    def __call__(self, program: itir.Program, *args: Any, **kwargs: Any) -> str:
        return self.formatter(program, *args, **kwargs)


def program_formatter(func: Callable[..., str]) -> ProgramFormatter:
    """
    Turn a function that formats a program as a string into a ProgramFormatter.

    Examples:
        >>> @program_formatter
        ... def format_foo(fencil: itir.Program, *args, **kwargs) -> str:
        ...     '''A very useless fencil formatter.'''
        ...     return "foo"

        >>> isinstance(format_foo, ProgramFormatter)
        True
    """

    return WrappedProgramFormatter(formatter=func)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import os
import pathlib
from typing import ContextManager, Protocol

import filelock

from gt4py.eve import utils as eve_utils


class _LockConstructor(Protocol):
    def __call__(self, path: str | os.PathLike, /) -> ContextManager: ...

    @property
    def __name__(self) -> str: ...


# Changing the lock is private API, but could be made public in the future
_lock_constructor: _LockConstructor = filelock.FileLock


def lock(directory: pathlib.Path | str, suffix: str = ".lock") -> ContextManager:
    """
    Create a lock for the given path.

    If the path is a directory, it must already exist.
    """
    directory = pathlib.Path(directory)

    # Identifier of the lock implementation to avoid conflicts
    # when switching between different implementations.
    identifier = f"{eve_utils.a10n(_lock_constructor.__module__)}_{eve_utils.a10n(_lock_constructor.__name__)}"

    if directory.is_dir():
        path = directory / f"{identifier}{suffix}"
    else:
        path = directory.with_suffix(f".{identifier}{suffix}")

    return _lock_constructor(str(path))



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import copyreg
import dataclasses
import functools
import inspect
import itertools
import types
from typing import (
    Any,
    Callable,
    ClassVar,
    Final,
    Optional,
    ParamSpec,
    Sequence,
    TypeAlias,
    TypeGuard,
    TypeVar,
    cast,
    overload,
)

from gt4py.eve import datamodels, utils as eve_utils


GT4PY_CLASS_METADATA_NS: Final[str] = "GT4PY_META"


def gt4py_metadata(**kwargs: Any) -> dict[str, dict[str, Any]]:
    """Helper function to store dataclass/datamodel field metadata within a GT4Py namespace."""
    return {GT4PY_CLASS_METADATA_NS: kwargs}


_StandardPickleState = None | dict[str, Any] | tuple[dict[str, Any] | None, dict[str, Any]]
_StandardGetStateMethod: TypeAlias = Callable[[object], _StandardPickleState]
_StandardSetStateMethod: TypeAlias = Callable[[object, _StandardPickleState], None]


@functools.cache
def _get_metadata_based_state_getstate(cls: type) -> _StandardGetStateMethod:
    """
    Helper function to make class-specific `__getstate__` method following the standard implementation.
    """
    if not isinstance(cls, type) or not (
        (is_dataclass := dataclasses.is_dataclass(cls)) or datamodels.is_datamodel(cls)
    ):
        raise TypeError(f"Expected a dataclass or datamodel type, got '{cls}'")

    if is_dataclass:
        field_metadata = {
            field.name: field.metadata.get(GT4PY_CLASS_METADATA_NS, None)
            for field in dataclasses.fields(cls)
        }
    else:
        field_metadata = {
            field.name: field.metadata.get(GT4PY_CLASS_METADATA_NS, None)
            for field in datamodels.get_fields(cls).values()
        }

    # To gather all slots we need to traverse the whole MRO not just the class itself.
    # We reuse the implementation of `copyreg._slotnames` to avoid code duplication and
    # potential bugs, even if it is not (unfortunately) part of the module's public API.
    class_slots = copyreg._slotnames(cls)  # type: ignore[attr-defined] # copyreg._slotnames is not recognized

    has_slots = len(class_slots) > 0
    has_dict = any("__dict__" in c.__dict__ for c in cls.__mro__)

    dict_names = []
    slot_names = []
    for name, metadata in field_metadata.items():
        if metadata and not metadata.get("pickle", True):
            continue
        if name in class_slots:
            slot_names.append(name)
        else:
            dict_names.append(name)

    # Comply with the default implementation of object.__getstate__() / object.__setstate__(state)
    if not (has_slots or has_dict):
        # for class instances without __dict__ nor __slots__: state = None.
        def __getstate__(self: object) -> _StandardPickleState:
            return None

    elif not has_slots:
        # for class instances with __dict__ and no __slots__: state = self.__dict__.
        def __getstate__(self: object) -> _StandardPickleState:
            return {name: getattr(self, name) for name in dict_names}

    elif not has_dict:
        # for class instances with __slots__ and no __dict__: state = (None, {slot.name: slot.value for all slots}).
        def __getstate__(self: object) -> _StandardPickleState:
            return (None, {name: getattr(self, name) for name in slot_names})

    else:
        # for class instances with __dict__ and __slots__: state = (self.__dict__, {slot.name: slot.value for all slots}).
        def __getstate__(self: object) -> _StandardPickleState:
            return (
                {name: getattr(self, name) for name in dict_names},
                {name: getattr(self, name) for name in slot_names},
            )

    return __getstate__


class MetadataBasedPickling:
    """
    Mixin for adding metadata-based pickling to dataclass-like objects.

    It uses the class field information to select only instance fields which
    are not marked with `pickle=False` in the 'GT4PY_META' metadata namespace.
    Individual fields can therefore opt out of pickling.
    For example: `foo = field(..., metadata=gt4py_metadata(pickle=False))`.
    """

    __slots__ = ()  # to avoid creation of __dict__ when not needed

    def __getstate__(self) -> _StandardPickleState:
        """
        Get the state of the object for pickling.

        It returns the same kind of arguments as the default `__getstate__`
        implementation, used by `pickle` (as documented in `object.__getstate__`,
        check: https://devdocs.io/python~3.14/library/pickle#object.__getstate__)
        """
        return _get_metadata_based_state_getstate(type(self))(self)  # type: ignore[arg-type]  # type(self) should be hashable

    # Note: we don't implement `__setstate__` as the output of our custom
    #  `__getstate__` implementation should be compatible with the default
    #  implementation.


class RecursionGuard:
    """
    Context manager to guard against inifinite recursion.

    >>> def foo(i):
    ...     with RecursionGuard(i):
    ...         if i % 2 == 0:
    ...             foo(i)
    ...     return i
    >>> foo(3)
    3
    >>> foo(2)  # doctest:+ELLIPSIS
    Traceback (most recent call last):
        ...
    gt4py.next.utils.RecursionGuard.RecursionDetected
    """

    guarded_objects: ClassVar[set[int]] = set()

    obj: Any

    class RecursionDetected(Exception):
        pass

    def __init__(self, obj: Any):
        self.obj = obj

    def __enter__(self) -> None:
        if id(self.obj) in self.guarded_objects:
            raise self.RecursionDetected()
        self.guarded_objects.add(id(self.obj))

    def __exit__(self, *exc: Any) -> None:
        self.guarded_objects.remove(id(self.obj))


_T = TypeVar("_T")
_P = ParamSpec("_P")
_R = TypeVar("_R")


def is_tuple_of(v: Any, t: type[_T]) -> TypeGuard[tuple[_T, ...]]:
    return isinstance(v, tuple) and all(isinstance(e, t) for e in v)


# TODO(havogt): remove flatten duplications in the whole codebase
def flatten_nested_tuple(
    value: tuple[
        _T | tuple, ...
    ],  # `_T` omitted on purpose as type of `value`, to properly deduce `_T` on the user-side
) -> tuple[_T, ...]:
    if isinstance(value, tuple):
        return sum((flatten_nested_tuple(v) for v in value), start=())  # type: ignore[arg-type] # cannot properly express nesting
    else:
        return (value,)


@overload
def tree_map(
    fun: Callable[_P, _R],
    *,
    collection_type: type | tuple[type, ...] = tuple,
    result_collection_constructor: Optional[Callable] = None,
    unpack: bool = False,
    with_path_arg: bool = False,
) -> Callable[..., _R | tuple[_R | tuple, ...]]: ...


@overload
def tree_map(
    *,
    collection_type: type | tuple[type, ...] = tuple,
    result_collection_constructor: Optional[Callable] = None,
    unpack: bool = False,
    with_path_arg: bool = False,
) -> Callable[
    [Callable[_P, _R]], Callable[..., Any]
]: ...  # TODO(havogt): typing of `result_collection_constructor` is too weak here


def tree_map(
    fun: Optional[Callable[_P, _R]] = None,
    *,
    collection_type: type | tuple[type, ...] = tuple,
    result_collection_constructor: Optional[Callable] = None,
    unpack: bool = False,
    with_path_arg: bool = False,
) -> Callable[..., _R | tuple[_R | tuple, ...]] | Callable[[Callable[_P, _R]], Callable[..., Any]]:
    """
    Apply `fun` to each entry of (possibly nested) collections (by default `tuple`s).

    Args:
        fun: Function to apply to each entry of the collection.
        collection_type: Type of the collection to be traversed. Can be a single type or a tuple of types.
        result_collection_constructor: Type of the collection to be returned. If `None` the same type as `collection_type` is used.
        unpack: Replicate tuple structure returned from `fun` to the mapped result, i.e. return
          tuple of result collections instead of result collections of tuples.
        with_path_arg: Pass the path to access the current element to `fun`.
    Examples:
        >>> tree_map(lambda x: x + 1)(((1, 2), 3))
        ((2, 3), 4)

        >>> tree_map(lambda x, y: x + y)(((1, 2), 3), ((4, 5), 6))
        ((5, 7), 9)

        >>> tree_map(collection_type=list)(lambda x: x + 1)([[1, 2], 3])
        [[2, 3], 4]

        >>> tree_map(
        ...     collection_type=(list, tuple),
        ...     result_collection_constructor=lambda value, elts: tuple(elts)
        ...     if isinstance(value, list)
        ...     else list(elts),
        ... )(lambda x: x + 1)([(1, 2), 3])
        ([2, 3], 4)

        >>> @tree_map
        ... def impl(x):
        ...     return x + 1
        >>> impl(((1, 2), 3))
        ((2, 3), 4)

        >>> @tree_map(with_path_arg=True)
        ... def impl(x, path: tuple[int, ...]):
        ...     path_str = "".join(f"[{i}]" for i in path)
        ...     return f"t{path_str} = {x}"
        >>> t = impl(((1, 2), 3))
        >>> t[0][0]
        't[0][0] = 1'
        >>> t[0][1]
        't[0][1] = 2'
        >>> t[1]
        't[1] = 3'

        >>> @tree_map(unpack=True)
        ... def impl(x):
        ...     return (x, x**2)
        >>> identity, squared = impl(((2, 3), 4))
        >>> identity
        ((2, 3), 4)
        >>> squared
        ((4, 9), 16)
    """

    if result_collection_constructor is None:
        if isinstance(collection_type, tuple):
            # Note: that doesn't mean `collection_type=tuple`, but e.g. `collection_type=(list, tuple)`
            raise TypeError(
                "tree_map() requires `result_collection_constructor` when `collection_type` is a tuple of types."
            )
        result_collection_constructor = lambda _, elts: collection_type(elts)  # noqa: E731 # because a lambda is clearer

    if fun:

        @functools.wraps(fun)
        def impl(*args: Any | tuple[Any | tuple, ...]) -> _R | tuple[_R | tuple, ...]:
            if isinstance(args[0], collection_type):
                non_path_args: Sequence[Any]
                if with_path_arg:
                    *non_path_args, path = args
                    args = (*non_path_args, tuple((*path, i) for i in range(len(args[0]))))
                else:
                    non_path_args = args

                assert all(
                    isinstance(arg, collection_type) and len(args[0]) == len(arg)
                    for arg in non_path_args
                )
                assert result_collection_constructor is not None
                ctor = functools.partial(result_collection_constructor, args[0])

                mapped = [impl(*arg) for arg in zip(*args)]
                if unpack:
                    return tuple(map(ctor, zip(*mapped)))
                else:
                    return ctor(mapped)

            return fun(  # type: ignore[call-arg]
                *cast(_P.args, args),  # type: ignore[valid-type]
            )  # mypy doesn't understand that `args` at this point is of type `_P.args`

        if with_path_arg:
            return lambda *args: impl(*args, ())
        else:
            return impl
    else:
        return functools.partial(
            tree_map,
            collection_type=collection_type,
            result_collection_constructor=result_collection_constructor,
            unpack=unpack,
            with_path_arg=with_path_arg,
        )


_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2")


def equalize_tuple_structure(
    d1: _T1, d2: _T2, *, fill_value: Any, bidirectional: bool = True
) -> tuple[_T1, _T2]:
    """
    Given two values or tuples thereof equalize their structure.

    If one of the arguments is a tuple the argument will be promoted to a tuple of the same
    structure.

    >>> equalize_tuple_structure((1,), (2, 3), fill_value=None) == (
    ...     (1, None),
    ...     (2, 3),
    ... )
    True

    >>> equalize_tuple_structure(1, (2, 3), fill_value=None) == (
    ...     (1, 1),
    ...     (2, 3),
    ... )
    True

    If `bidirectional` is `False` only the second argument is equalized and an error is raised if
    the first argument would require promotion.
    """
    d1_is_tuple = isinstance(d1, tuple)
    d2_is_tuple = isinstance(d2, tuple)
    if not d1_is_tuple and d2_is_tuple:
        if not bidirectional:
            raise ValueError(f"Expected `{d2!s}` to have same structure as `{d1!s}`.")
        return equalize_tuple_structure(
            (d1,) * len(d2),  # type: ignore[arg-type] # ensured by condition above
            d2,
            fill_value=fill_value,
            bidirectional=bidirectional,
        )
    if d1_is_tuple and not d2_is_tuple:
        return equalize_tuple_structure(
            d1,
            (d2,) * len(d1),  # type: ignore[arg-type] # ensured by condition above
            fill_value=fill_value,
            bidirectional=bidirectional,
        )
    if d1_is_tuple and d2_is_tuple:
        if not bidirectional and len(d1) < len(d2):  # type: ignore[arg-type] # ensured by condition above
            raise ValueError(f"Expected `{d2!s}` to be longer than or equal to `{d1!s}`.")
        return tuple(  # type: ignore[return-value] # mypy not smart enough
            zip(
                *(
                    equalize_tuple_structure(
                        el1, el2, fill_value=fill_value, bidirectional=bidirectional
                    )
                    for el1, el2 in itertools.zip_longest(d1, d2, fillvalue=fill_value)  # type: ignore[call-overload] # d1, d2 are tuples
                )
            )
        )
    return d1, d2


def make_args_canonicalizer(
    signature: inspect.Signature,
    *,
    name: str | None = None,
) -> Callable[..., tuple[tuple, dict[str, Any]]]:
    """
    Create a call arguments canonicalizer function from a given signature.

    The canonicalizer returns the call arguments in a tuple with the
    following two items:

        - a tuple of values containing all the positional parameters
        - a dictionary of parameter names to values for all keyword-only parameters

    Args:
        signature: The signature for which to create the canonicalizer.

    Keyword Args:
        name: Name of the function for which the canonicalizer is created.

    """

    params = []
    pos_args = []
    key_args = []
    var_pos_arg: str | None = None
    var_key_arg: str | None = None

    for key, param in signature.parameters.items():
        match param.kind:
            case inspect.Parameter.POSITIONAL_ONLY | inspect.Parameter.POSITIONAL_OR_KEYWORD:
                pos_args.append(f"{key}")
            case inspect.Parameter.VAR_POSITIONAL:
                var_pos_arg = key
                pos_args.append(f"*{key}")
            case inspect.Parameter.KEYWORD_ONLY:
                key_args.append(f"'{key}': {key}")
            case inspect.Parameter.VAR_KEYWORD:
                var_key_arg = key
                key_args.append(f"**{key}")

        # Remove annotations to avoid issues in case `str(annotation)`
        # produces incorrect python expressions
        params.append(param.replace(annotation=param.empty))

    canonicalizer_signature = inspect.Signature(parameters=params)

    if len(pos_args) == 1:
        if var_pos_arg:
            # If there is only an '*args' parameter, we can just return it directly
            pos_args_tuple_expr = var_pos_arg
        else:
            pos_args_tuple_expr = f"({pos_args[0]},)"  # Add trailing comma to make it a tuple
    else:
        # In the regular case, assemble the output tuple with all positional arguments
        pos_args_tuple_expr = f"({str.join(', ', pos_args)})"

    if len(key_args) == 1 and var_key_arg:
        # If there is only a '**kwargs' parameter, we can just return it directly
        key_args_dict_expr = var_key_arg
    else:
        # In the regular case, assemble the output dictionary with all keyword arguments
        key_args_dict_expr = f"{{ {str.join(', ', key_args)} }}"

    canonicalize_func_name = f"canonicalize_args_for_{name}" if name else "canonicalize_args"
    canonicalizer_src = f"""
def {canonicalize_func_name}{canonicalizer_signature!s}:
    return {pos_args_tuple_expr}, {key_args_dict_expr}
"""
    namespace: dict[str, Any] = {}
    exec(canonicalizer_src, namespace)
    canonicalizer = namespace[canonicalize_func_name]

    return cast(Callable[..., tuple[tuple, dict[str, Any]]], canonicalizer)


@functools.cache
def make_args_canonicalizer_for_function(
    func: types.FunctionType,
) -> Callable[..., tuple[tuple, dict[str, Any]]]:
    return make_args_canonicalizer(inspect.signature(func), name=func.__name__)


def canonicalize_call_args(
    func: Callable,
    /,
    args: tuple,
    kwargs: dict[str, Any],
) -> tuple[tuple, dict[str, Any]]:
    """
    Canonicalize call arguments for a given function.

    Args:
        func: The function for which to canonicalize the call arguments.
        args: Positional arguments.
        kwargs: Keyword arguments.

    Returns:
        A tuple of positional arguments and a dictionary with keyword arguments.

    Note:
        This is a convenience wrapper around `make_args_canonicalizer_for_function`.
    """

    return make_args_canonicalizer_for_function(func)(*args, **kwargs)


class IDGeneratorPool(eve_utils.CustomDefaultDictBase):
    """
    Utility for providing unique IDs for each prefix.

    The use-case for this implementation is to provide a single IDGeneratorPool for the whole
    transformation pipeline, which ensure unique names (prefix+counter) across all transformations.

    The reason for this implementation of a single counter per prefix (over a global increasing counter)
    is that parts of the IR might still be stable when comparing different versions of a transformation.

    >>> uids = IDGeneratorPool()
    >>> next(uids["foo"])
    'foo_0'
    >>> next(uids["foo"])
    'foo_1'
    >>> next(uids["bar"])
    'bar_0'
    """

    def value_factory(self, prefix: str) -> eve_utils.SequentialIDGenerator:
        return eve_utils.SequentialIDGenerator(prefix=prefix)



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import dataclasses
from typing import Optional, Sequence, TypeAlias

import dace
from dace import subsets as dace_subsets

from gt4py import eve
from gt4py.eve.extended_typing import MaybeNestedInTuple
from gt4py.next import common as gtx_common
from gt4py.next.iterator import ir as gtir
from gt4py.next.iterator.ir_utils import common_pattern_matcher as cpm, domain_utils
from gt4py.next.program_processors.runners.dace.lowering import gtir_to_sdfg_utils


@dataclasses.dataclass(frozen=True)
class FieldopDomainRange:
    """
    Represents the range of a field operator domain in one dimension.

    It contains 3 elements:
        dim: dimension definition
        start: symbolic expression for lower bound (inclusive)
        stop: symbolic expression for upper bound (exclusive)
    """

    dim: gtx_common.Dimension
    start: dace.symbolic.SymbolicType
    stop: dace.symbolic.SymbolicType


FieldopDomain: TypeAlias = list[FieldopDomainRange]
"""Domain of a field operator represented as a list of `FieldopDomainRange` for each dimension."""


def get_field_domain(domain: domain_utils.SymbolicDomain) -> FieldopDomain:
    """
    Visits the domain of a field operator and returns a list of dimensions and
    the corresponding lower and upper bounds. The returned lower bound is inclusive,
    the upper bound is exclusive, i.e. `[lower_bound, upper_bound[`

    Note that the doman dimensions are sorted in gt4py canonical order.
    """
    return [
        FieldopDomainRange(
            dim,
            gtir_to_sdfg_utils.get_symbolic(domain.ranges[dim].start),
            gtir_to_sdfg_utils.get_symbolic(domain.ranges[dim].stop),
        )
        for dim in gtx_common.order_dimensions(domain.ranges.keys())
    ]


TargetDomain: TypeAlias = MaybeNestedInTuple[domain_utils.SymbolicDomain]
"""Symbolic domain which defines the range to write in the target field.

For tuple output in fieldview, `TargetDomain` is a tree-like tuple of symbolic domains.
"""


def extract_target_domain(node: gtir.Expr) -> TargetDomain:
    """
    Visit a GTIR domain expression and construct a `TargetDomain` symbolic domain.

    We use a visitor class to extract the tree-like structure for (nested) tuple of domains.
    """

    class TargetDomainParser(eve.visitors.NodeTranslator):
        def visit_FunCall(self, node: gtir.FunCall) -> TargetDomain:
            if cpm.is_call_to(node, "make_tuple"):
                return tuple(self.visit(arg) for arg in node.args)
            else:
                return domain_utils.SymbolicDomain.from_expr(node)

    return TargetDomainParser().visit(node)


def get_element_subset(
    dims: Sequence[gtx_common.Dimension], origin: Optional[Sequence[dace.symbolic.SymExpr]]
) -> dace_subsets.Range:
    """
    Helper function to construct the Memlet subset to access an element in the field
    domain, applying an optional origin in each dimension as start index.

    Args:
        dims: The field dimensions.
        origin: The domain start index in each dimension. If set to `None`, assume all zeros.

    Returns:
        Range containing the indices for field access in dace arrays. Although it is a
        `dace.subsets.Range` object, it only encodes a single value, thus only the
        lower bound has any significance.
    """
    # TODO(phimuell, edopao): Think about if this function should return `list[SymExpr]`.
    assert len(dims) != 0
    index_variables = [
        dace.symbolic.pystr_to_symbolic(gtir_to_sdfg_utils.get_map_variable(dim)) for dim in dims
    ]
    origin = [0] * len(index_variables) if origin is None else origin
    start_values = [
        index - start_index for index, start_index in zip(index_variables, origin, strict=True)
    ]
    return dace_subsets.Range([(s, s, 1) for s in start_values])


def get_field_layout(
    field_domain: FieldopDomain,
) -> tuple[list[gtx_common.Dimension], list[dace.symbolic.SymExpr], list[dace.symbolic.SymExpr]]:
    """
    Parse the field operator domain and generate the shape of the result field.

    Note that this function also ensures that the array shape computed from the
    domain range is non-negative. A negative shape can occur in concat_where
    expressions, where it can happen that 'stop' value is smaller than 'start'.
    Also note that this _strange_ domain with 'start' > 'stop' is usually propagated
    to the input arguments of a concat_where expression (the child nodes), thus
    the lowering of regular field operators also needs to apply the sanity check
    in order to avoid allocation of temporary fields with negative size.

    Args:
        field_domain: The field operator domain.

    Returns:
        A tuple of three lists containing:
            - the domain dimensions
            - the domain origin, that is the start indices in all dimensions
            - the domain size in each dimension
    """
    if len(field_domain) == 0:
        return [], [], []
    domain_dims = [domain_range.dim for domain_range in field_domain]
    domain_origin = [domain_range.start for domain_range in field_domain]
    domain_shape = [
        dace.symbolic.pystr_to_symbolic(f"max(0, {domain_range.stop - domain_range.start})")
        for domain_range in field_domain
    ]
    return domain_dims, domain_origin, domain_shape



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

# Extracted from definitions.py to avoid `bool` to shadow the python built-in `bool` type in that file.

import numpy as np


# TODO(havogt): change to `np.bool` once we drop numpy 1.x support
bool = np.bool_  # noqa: A001

int8 = np.int8
int16 = np.int16
int32 = np.int32
int64 = np.int64

uint8 = np.uint8
uint16 = np.uint16
uint32 = np.uint32
uint64 = np.uint64

float32 = np.float32
float64 = np.float64

# This is used to map between dtypes of different array namespaces, as only the name is defined in the array API.
# We don't use `<type>.__name__` for the mapping to be on the safe side.
type_to_name = {
    globals()[name]: name
    for name in (
        "bool",
        "int8",
        "int16",
        "int32",
        "int64",
        "uint8",
        "uint16",
        "uint32",
        "uint64",
        "float32",
        "float64",
    )
}



# GT4Py - GridTools Framework
#
# Copyright (c) 2014-2024, ETH Zurich
# All rights reserved.
#
# Please, refer to the LICENSE file in the root directory.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import Callable, ClassVar, Optional, Union

from gt4py.eve import Coerced, SymbolName, datamodels
from gt4py.eve.traits import SymbolTableTrait, ValidatedSymbolTableTrait
from gt4py.next import common
from gt4py.next.iterator import builtins
from gt4py.next.program_processors.codegens.gtfn.gtfn_im_ir import ImperativeFunctionDefinition
from gt4py.next.program_processors.codegens.gtfn.gtfn_ir_common import Expr, Node, Sym, SymRef


class UnaryExpr(Expr):
    op: str
    expr: Expr


class BinaryExpr(Expr):
    op: str
    lhs: Expr
    rhs: Expr


class TernaryExpr(Expr):
    cond: Expr
    true_expr: Expr
    false_expr: Expr


class CastExpr(Expr):
    obj_expr: Expr
    new_dtype: SymRef


class Literal(Expr):
    value: str
    type: str


class IntegralConstant(Expr):
    value: int  # generalize to other types if needed


class OffsetLiteral(Expr):
    value: Union[int, str]


class Lambda(Expr, SymbolTableTrait):
    params: list[Sym]
    expr: Expr


class FunCall(Expr):
    fun: Expr  # VType[Callable]
    args: list[Expr]


class FunctionDefinition(Node, SymbolTableTrait):
    id: Coerced[SymbolName]
    params: list[Sym]
    expr: Expr


class ScanPassDefinition(Node, SymbolTableTrait):
    id: Coerced[SymbolName]
    params: list[Sym]
    expr: Expr
    forward: bool
    projector: Optional[Lambda]


class TaggedValues(Node):
    tags: list[Expr]
    values: list[Expr]


class CartesianDomain(Node):
    tagged_sizes: TaggedValues
    tagged_offsets: TaggedValues


class UnstructuredDomain(Node):
    tagged_sizes: TaggedValues
    tagged_offsets: TaggedValues
    connectivities: list[SymRef]  # SymRef to offset declaration


class Backend(Node):
    domain: Union[SymRef, CartesianDomain, UnstructuredDomain]


def _is_tuple_expr_of(pred: Callable[[Expr], bool], expr: Expr) -> bool:
    if (
        isinstance(expr, FunCall)
        and isinstance(expr.fun, SymRef)
        and expr.fun.id == "tuple_get"
        and len(expr.args) == 2
        and _is_tuple_expr_of(pred, expr.args[1])
    ):
        return True
    if (
        isinstance(expr, FunCall)
        and isinstance(expr.fun, SymRef)
        and expr.fun.id == "make_tuple"
        and all(_is_tuple_expr_of(pred, arg) for arg in expr.args)
    ):
        return True
    return pred(expr)


class SidComposite(Expr):
    values: list[Expr]

    @datamodels.validator("values")
    def _values_validator(
        self: datamodels.DataModelTP, attribute: datamodels.Attribute, value: list[Expr]
    ) -> None:
        if not all(
            isinstance(el, (SidFromScalar, SidComposite))
            or _is_tuple_expr_of(
                lambda expr: isinstance(expr, (SymRef, Literal))
                or (isinstance(expr, FunCall) and expr.fun == SymRef(id="index")),
                el,
            )
            for el in value
        ):
            raise ValueError(
                "Only 'SymRef', 'Literal', tuple expr thereof, 'SidFromScalar', or 'SidComposite' allowed."
            )


def _might_be_scalar_expr(expr: Expr) -> bool:
    if isinstance(expr, BinaryExpr):
        return all(_is_tuple_expr_of(_might_be_scalar_expr, arg) for arg in (expr.lhs, expr.rhs))
    if isinstance(expr, UnaryExpr):
        return _is_tuple_expr_of(_might_be_scalar_expr, expr.expr)
    if (
        isinstance(expr, FunCall)
        and isinstance(expr.fun, SymRef)
        and expr.fun.id in ARITHMETIC_BUILTINS
    ):
        return all(_might_be_scalar_expr(arg) for arg in expr.args)
    if isinstance(expr, CastExpr):
        return _might_be_scalar_expr(expr.obj_expr)
    if _is_tuple_expr_of(lambda e: isinstance(e, (SymRef, Literal)), expr):
        return True
    return False


class SidFromScalar(Expr):
    arg: Expr

    @datamodels.validator("arg")
    def _arg_validator(
        self: datamodels.DataModelTP, attribute: datamodels.Attribute, value: Expr
    ) -> None:
        if not _might_be_scalar_expr(value):
            raise ValueError(
                "Only 'SymRef', 'Literal', arithmetic op or tuple expr thereof allowed."
            )


class Stmt(Node):
    pass


class StencilExecution(Stmt):
    backend: Backend
    stencil: SymRef
    output: Union[SymRef, SidComposite]
    inputs: list[Union[SymRef, SidComposite, SidFromScalar, FunCall]]

    @datamodels.validator("inputs")
    def _arg_validator(
        self: datamodels.DataModelTP, attribute: datamodels.Attribute, inputs: list[Expr]
    ) -> None:
        for inp in inputs:
            if not _is_tuple_expr_of(
                lambda expr: isinstance(expr, (SymRef, SidComposite, SidFromScalar))
                or (
                    isinstance(expr, FunCall)
                    and isinstance(expr.fun, SymRef)
                    and expr.fun.id == "index"
                ),
                inp,
            ):
                raise ValueError(
                    "Only 'SymRef', 'SidComposite', 'SidFromScalar', 'index' call or tuple expr thereof allowed."
                )


class Scan(Node):
    function: SymRef
    output: int
    inputs: list[int]
    init: Expr


class ScanExecution(Stmt):
    backend: Backend
    scans: list[Scan]
    args: list[Expr]
    axis: SymRef


class IfStmt(Stmt):
    cond: Expr
    true_branch: list[Stmt]
    false_branch: list[Stmt]


class TemporaryAllocation(Node):
    id: SymbolName
    dtype: str
    domain: Union[SymRef, CartesianDomain, UnstructuredDomain]


GTFN_BUILTINS = [
    "deref",
    "shift",
    "make_tuple",
    "tuple_get",
    "can_deref",
    "cartesian_domain",
    "unstructured_domain",
    "get_domain_range",
    "named_range",
    "reduce",
    "index",
]
ARITHMETIC_BUILTINS = builtins.ARITHMETIC_BUILTINS
TYPEBUILTINS = builtins.TYPE_BUILTINS

BUILTINS = {*GTFN_BUILTINS, *ARITHMETIC_BUILTINS, *TYPEBUILTINS}


class TagDefinition(Node):
    name: Sym
    alias: Optional[Union[str, SymRef]] = None


class Program(Node, ValidatedSymbolTableTrait):
    id: SymbolName
    params: list[Sym]
    function_definitions: list[
        Union[FunctionDefinition, ScanPassDefinition, ImperativeFunctionDefinition]
    ]
    executions: list[Stmt]
    offset_definitions: list[TagDefinition]
    grid_type: common.GridType
    temporaries: list[TemporaryAllocation]

    _NODE_SYMBOLS_: ClassVar[list[Sym]] = [Sym(id=name) for name in BUILTINS]
